/*
 * FILE:     tfrc_rcv.c
 * AUTHOR:   Ladan Gharai <ladan@isi.edu>
 * MODIFIED: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2002-2005 USC Information Sciences Institute
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.9 $
 * $Date: 2005/07/28 00:54:49 $
 *
 *----------------------------------------------------------------------------
 *
 * This file implements the receiver functionalites of a TFRC receiver per 
 * RFC 3448. This implementation assumes the use of send time stamps with
 * millisecond granularity (and not quad time counters).  Sec x.x.x in this 
 * file always refers to Sections of RFC 3448 - LG. 
 *
 *----------------------------------------------------------------------------
 */

#include "config.h"
#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <sys/time.h>
#include <netinet/in.h>
#include "tv.h"
#include "tfrc_rcv.h"


#define TFRC_MAGIC	 0xbaef03b7	                     /* For debugging */
#define RTP_SEQ_MOD	    0x10000
#define MAX_MISORDER	        200
#define TFRC_MAX_HISTORY      10000
#define TFRC_NUM_INTERVALS        8                    /* number of intervals */
#define TFRC_MAX_DROPOUT       5000        /* perhaps this should be smaller? */
                                           /* anyway > MAX_DROPOUT packets are*/
                                           /* lost go back to slow start      */
#define TFRC_RTT_INIT       1000000               /* intial RTT value  = 1sec */
#define TFRC_PSIZE_INIT        1470
/******************************************************************************/
/*                                                                            */
/******************************************************************************/
struct tfrc_history {
        uint32_t seq;
        uint32_t ts;
}; 

/*
 * TFRC reciever variables
 */
struct tfrc_rcv {
        double          p;
	uint32_t        lsend_ts;        /* timestamp of last packet received */
	uint32_t        x_recv;
        uint32_t        rcv_rtt;    /* rtt receiver extracts from rtp packets */
	uint32_t	pckt_size;
        uint32_t        sent_srx;         /* time last TFRC feedback was sent */
	uint32_t        last_pckt_t;               /* rtp packet receive time */
	uint32_t	num_pckts;   /* no packets received since last feedbck*/
	struct timeval  feedbk_timer;

        uint16_t 	last_seq;       
        uint32_t 	ext_last_seq;
        int      	last_seq_jj;
	uint16_t	interval_count;   
        double          W_tot;
        double          weight[TFRC_NUM_INTERVALS+5]; /* weights for p RFC3448 */
        int             jj;       /* index for arrival -1<=i<=TFRC_MAX_HISTORY */
        int             ii;                    /* index for loss    -1<=ii<=10 */
        int             ooo;                 /* number of out of order packets */
        int             cycles;     /* number of times seq number cycles 65535 */
	int             loss_count;
	int             total_pckts;
	struct 
	  tfrc_history  loss[TFRC_MAX_HISTORY],
        	        arrival[TFRC_MAX_HISTORY];
        uint32_t        magic;                               /* For debugging */
};

int first_pckt = 0;                 /* indicates arrival of first data packet */
/******************************************************************************/
/*									      */
/******************************************************************************/
static void
validate_tfrc_rcvr (struct tfrc_rcv *tfrc)
{
	assert(tfrc!=NULL);
	assert(tfrc->magic == TFRC_MAGIC);
}


/* use reverse of Matt Mathis formula with c=1.22*/
static double
compute_init_p ( uint32_t RTT, uint16_t s, uint32_t rate ) {
        double rtt, tmp=0;

        assert (RTT!=0);
	assert (rate!=0);
        
        /* convert RTT from micro-sec to sec */
        rtt = ((double) RTT) / 1000000.0;
 
        tmp = (s *s * 1.44)/(rtt*rtt*rate*rate);
        return (tmp);
}


static int
set_zero(struct tfrc_rcv *tfrc, int first, int last, uint16_t u)
{
	int i, count = 0;;

	assert((first >= 0) && (first < TFRC_MAX_HISTORY));
	assert((last >= 0) && (last < TFRC_MAX_HISTORY));

	if (first == last)
		return (0);
	if ((first + 1) % TFRC_MAX_HISTORY == last)
		return (1);

	if (first < last) {
		for (i = first + 1; i < last; i++) {
			tfrc->arrival[i].seq = 0;
			tfrc->arrival[i].ts = 0;
			count++;
		}
	} else {
		for (i = first + 1; i < TFRC_MAX_HISTORY; i++) {
			tfrc->arrival[i].seq = 0;
			tfrc->arrival[i].ts = 0;
			count++;
		}
		for (i = 0; i < last; i++) {
			tfrc->arrival[i].seq = 0;
			tfrc->arrival[i].ts = 0;
			count++;
		}
	}
	assert(count == (u - 1));
	return (count);
}


static int
arrived(struct tfrc_rcv *tfrc, int first, int last)
{
	int i, count = 0;

	assert((first >= 0) && (first < TFRC_MAX_HISTORY));
	assert((last >= 0) && (last < TFRC_MAX_HISTORY));

	if (first == last)
		return (0);
	if ((first + 1) % TFRC_MAX_HISTORY == last)
		return (0);

	if (first < last) {
		for (i = first; i <= last; i++) {
			if (tfrc->arrival[i].seq != 0)
				count++;
		}
	} else {
		for (i = first; i < TFRC_MAX_HISTORY; i++) {
			if (tfrc->arrival[i].seq != 0)
				count++;
		}
		for (i = 0; i <= last; i++) {
			if (tfrc->arrival[i].seq != 0)
				count++;
		}
	}
	return (count);
}


static int
record_loss(struct tfrc_rcv *tfrc, uint32_t s1, uint32_t s2, uint32_t ts1, uint32_t ts2)
{
	/* Mark all packets between s1 (which arrived at time ts1) and */
	/* s2 (which arrived at time ts2) as lost.                     */
	int 		i;
	uint32_t 	est;
	uint32_t 	seq = s1 + 1;

	est = ts1 + (ts2 - ts1) * ((seq - s1) / (s2 - seq));

	tfrc->loss_count++;
	if (tfrc->ii <= -1) {
		/* first loss! */
		tfrc->ii++;
		tfrc->loss[tfrc->ii].seq = seq;
		tfrc->loss[tfrc->ii].ts = est;
		return 1;
	}

	/* not a new event */
	if (est - tfrc->loss[tfrc->ii].ts <= tfrc->rcv_rtt) {	
		return 0;
	}

	tfrc->interval_count++;
	if (tfrc->ii > (TFRC_NUM_INTERVALS + 1)) {
		printf("how did this happen?\n");
	}

	if (tfrc->ii >= (TFRC_NUM_INTERVALS + 1)) {	/* shift */
		for (i = 0; i < (TFRC_NUM_INTERVALS + 1); i++) {
			tfrc->loss[i].seq = tfrc->loss[i + 1].seq;
			tfrc->loss[i].ts = tfrc->loss[i + 1].ts;
		}
		tfrc->ii = TFRC_NUM_INTERVALS;
	}

	tfrc->ii++;
	tfrc->loss[tfrc->ii].seq = seq;
	tfrc->loss[tfrc->ii].ts = est;
	return 1;
}


/******************************************************************************/
/*                                                                            */
/*  External API follows...						      */
/*                                                                            */
/******************************************************************************/

uint32_t
tfrc_get_rcvr_rtt(struct tfrc_rcv *tfrc)
{
	assert (tfrc!=NULL);
	return tfrc->rcv_rtt;
}


void
tfrc_set_rcvr_psize (struct tfrc_rcv *tfrc, uint32_t a)
{
	assert (tfrc!=NULL);
	if (tfrc)
		tfrc->pckt_size = a;
}


void
tfrc_set_p (struct tfrc_rcv *tfrc, double p)
{
	assert (tfrc!=NULL);
        tfrc->p = p;
}


double
tfrc_get_p (struct tfrc_rcv *tfrc)
{
	assert (tfrc!=NULL);
       	return (tfrc->p);
}


double
tfrc_compute_p(struct tfrc_rcv *tfrc)
{
        int             i;
        uint32_t        temp, I_tot0 = 0, I_tot1 = 0, I_tot = 0;
        double          p,
                        I_mean;
	static double init_p=0;
        static int my_ii=0;


	assert (tfrc!=NULL);

        if (tfrc->ii ==-1)
        	/* no loss occured yet */
        	return 0;
        else 
        	if (tfrc->ii == 0) {
			if(init_p==0 ) {
                		/* this is the first loss interval compute */
                        	/* init_p */
                       		init_p = compute_init_p (tfrc->rcv_rtt,
                                                tfrc->pckt_size,
                                                tfrc->x_recv );
                       		printf ("N:0 %8.6f x:%dkbps  rtt:%d\n",
                                     init_p, (tfrc->x_recv*8)/1000, tfrc->rcv_rtt);
                	}
                        return init_p;
		}

	for (i = 0;  i < tfrc->ii; i++) {
                temp = tfrc->loss[i + 1].seq - tfrc->loss[i].seq;
                I_tot0 = I_tot0 + temp * tfrc->weight[i];
		if (i >= 1) {
                        I_tot1 = I_tot1 + temp * tfrc->weight[i];
                }
        }
        I_tot1 = I_tot1 + (tfrc->arrival[tfrc->jj].seq - tfrc->loss[tfrc->ii].seq) 
						* tfrc->weight[tfrc->ii];

        I_tot = (I_tot1 > I_tot0) ? I_tot1 : I_tot0;
        I_mean = I_tot / tfrc->W_tot;
        p = 1 / I_mean;

 	//if (tfrc->ii<TFRC_NUM_INTERVALS) 
		//printf ("\nN:%d %8.6f rtt:%d\n", tfrc->ii, p, tfrc->rcv_rtt);
        return p;
}


/*
 * return true if its time to send feedback AND reset the timer
 */
uint16_t
tfrc_send_rcvr_feedback (struct tfrc_rcv *tfrc, struct timeval now) 
{
	assert (tfrc!=NULL);

	if (tv_gt (now, tfrc->feedbk_timer) || first_pckt){
		/* reset the timer */
		tv_add_usec (&now, tfrc->rcv_rtt);
		tfrc->feedbk_timer.tv_sec = now.tv_sec;
                tfrc->feedbk_timer.tv_usec = now.tv_usec;

		if (tfrc->num_pckts>0 || first_pckt) 
			return (1);
		else
			return (0);

	}
	else
		return (0);
}

        
void
tfrc_format_feedback ( struct tfrc_rcv *tfrc,  uint32_t now,
                                uint32_t *tdelay, uint32_t *x_recv, uint32_t *send_ts, uint32_t *p)
{
        uint32_t my_tdelay = 0,
                 my_xrecv = 0;
  
        assert (tfrc!=NULL);
	if (first_pckt) {
		/* first packet - perpare feedback per Sec 6.3 */
		assert (tfrc->p==0);
		assert (tfrc->x_recv==0);
	}
	else {
        	my_xrecv = tfrc->num_pckts*tfrc->pckt_size;
        	my_xrecv = my_xrecv / ((now - tfrc->sent_srx)/1000000.0) ;
	}
        my_tdelay = now - tfrc->last_pckt_t;
                
        *tdelay = my_tdelay;
        *send_ts = tfrc->lsend_ts;
        *x_recv = my_xrecv;
        *p = tfrc->p * 4294967295u;
                
        tfrc->num_pckts = 0;
        tfrc->sent_srx = now;
        tfrc->x_recv=my_xrecv;
}



uint16_t
tfrc_record_arrival(struct tfrc_rcv *tfrc,  uint32_t now, 
			uint16_t seq, uint32_t send_ts, uint16_t rtt_flag, uint32_t rtt )
{
	int 		kk, 
			inc, 
			last_jj,
			new_loss_event=0;
	uint16_t 	udelta;
	uint32_t	ext_seq;
 
	assert (tfrc!=NULL);
 
	/* squirrel away some info */ 
        tfrc->last_pckt_t = now;
	tfrc->lsend_ts = send_ts;
	tfrc->num_pckts++;
        if (rtt_flag) {
		/* new RTT arrived in RTP packet */
		tfrc->rcv_rtt = rtt;
        }

	if (tfrc->jj == -1) {
		/* first packet arrival */
		tfrc->jj = 0;
		tfrc->last_seq = seq;
		tfrc->ext_last_seq = seq;
		tfrc->last_seq_jj = 0;
		tfrc->arrival[tfrc->jj].seq = seq;
		tfrc->arrival[tfrc->jj].ts = now; 
		first_pckt = 1;
		return 0;
	}
	first_pckt=0;

	udelta = seq - tfrc->last_seq;
	tfrc->total_pckts++;
	if (udelta < TFRC_MAX_DROPOUT) {
		/* in order, with permissible gap */
		if (seq < tfrc->last_seq) {
			tfrc->cycles++;
		}
		/* record arrival */
		last_jj = tfrc->jj;
		tfrc->jj = (tfrc->jj + udelta) % TFRC_MAX_HISTORY;
		set_zero(tfrc, last_jj, tfrc->jj, udelta);
		ext_seq = seq + tfrc->cycles * RTP_SEQ_MOD;
		tfrc->last_seq = seq;
		tfrc->arrival[tfrc->jj].seq = ext_seq;
		tfrc->arrival[tfrc->jj].ts = now;

		if ((ext_seq - tfrc->ext_last_seq) == 1) {
			/* We got two consecutive packets, no loss */
			tfrc->ext_last_seq = ext_seq;
			tfrc->last_seq_jj = tfrc->jj;
		} else {
			/* Sequence number jumped, we've missed a packet for */
			/* some reason, ex : arrived()=2 for  3654 and  3656 */
			if (arrived(tfrc, tfrc->last_seq_jj, tfrc->jj) >= 2) {
				new_loss_event = record_loss(tfrc, tfrc->ext_last_seq, ext_seq,
					    tfrc->arrival[tfrc->last_seq_jj].ts, now);
				tfrc->ext_last_seq = ext_seq;
				tfrc->last_seq_jj = tfrc->jj;
			}
			/* the else here can be useful for emualting TCP */
		}
	} else if (udelta <= RTP_SEQ_MOD - MAX_MISORDER) {
		printf("\ndelta:%d, seq:%u last_seq:%u", udelta, seq, tfrc->arrival[tfrc->jj].seq);
	} else {
		/* duplicate or reordered packet */
		printf ("===");
		tfrc->ooo++;
		ext_seq = seq + tfrc->cycles * RTP_SEQ_MOD;
		if (ext_seq > tfrc->ext_last_seq) {
			inc = ext_seq - tfrc->arrival[tfrc->jj].seq;

			kk = (tfrc->jj + inc) % TFRC_MAX_HISTORY;
			if (tfrc->arrival[kk].seq == 0) {
				tfrc->arrival[kk].seq = ext_seq;
				/* NOT the best interpolation */
				tfrc->arrival[kk].ts  = (tfrc->arrival[tfrc->last_seq_jj].ts + now) / 2;	
			}
			while (tfrc->arrival[tfrc->last_seq_jj + 1].seq != 0
			       && tfrc->last_seq_jj < tfrc->jj) {
				tfrc->last_seq_jj = (tfrc->last_seq_jj + 1) % TFRC_MAX_HISTORY;
			}
			tfrc->ext_last_seq = tfrc->arrival[tfrc->last_seq_jj].seq;
		}
	}
	return new_loss_event;
}


struct tfrc_rcv *
tfrc_rcvr_init(struct timeval tnow)
{
	int i;
	struct tfrc_rcv *tfrc;
	uint32_t now = timeval_to_usec (tnow);

	tfrc= (struct tfrc_rcv *) malloc (sizeof (struct tfrc_rcv));
	tfrc->p        = 0;
	tfrc->lsend_ts = 0;
	tfrc->x_recv   = 0;
	tfrc->rcv_rtt  = TFRC_RTT_INIT;
	tfrc->pckt_size= TFRC_PSIZE_INIT;
	tfrc->sent_srx = now; 
	tfrc->last_pckt_t = 0;
	tfrc->num_pckts = 0;

        tfrc->feedbk_timer.tv_sec  = tnow.tv_sec;
        tfrc->feedbk_timer.tv_usec = tnow.tv_usec;
	tv_add (&(tfrc->feedbk_timer), tfrc->rcv_rtt);

	tfrc->last_seq = 0;
        tfrc->ext_last_seq= 0;
        tfrc->last_seq_jj=0;
	tfrc->interval_count = 0;
	tfrc->W_tot     = 0;
	tfrc->weight[0] = 0.2;
	tfrc->weight[1] = 0.4;
	tfrc->weight[2] = 0.6;
	tfrc->weight[3] = 0.8;
	tfrc->weight[4] = 1.0;
	tfrc->weight[5] = 1.0;
	tfrc->weight[6] = 1.0;
	tfrc->weight[7] = 1.0;
	tfrc->weight[8] = 1.0;
        tfrc->ii       = -1;
        tfrc->jj       = -1;
	tfrc->ooo         = 0;
	tfrc->cycles      = 0;
        tfrc->loss_count  = 0;
	tfrc->total_pckts = 0;

	for (i = 0; i < TFRC_NUM_INTERVALS; i++) {
		tfrc->W_tot = tfrc->W_tot + tfrc->weight[i];
	}
	for (i = 0; i < TFRC_MAX_HISTORY; i++) {
		tfrc->arrival[i].seq = 0;
		tfrc->arrival[i].ts  = 0;
		tfrc->loss[i].seq = 0;
		tfrc->loss[i].ts  = 0;
	}
	tfrc->magic = TFRC_MAGIC;
	return (tfrc);
}
