/*
 *
 * FILE:     tfrc_snd.c
 * AUTHOR:   Ladan Gharai <ladan@isi.edu>
 *
 * Copyright (c) 2002-2005 USC Information Sciences Institute
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.9 $
 * $Date: 2005/07/27 00:10:11 $
 *
 *----------------------------------------------------------------------------
 *
 * This file implements the sender functionalites of a TFRC receiver per 
 * RFC 3448.  This implementation assumes the use of send time stamps with
 * millisecond granularity (and not quad time counters).  Sec x.x.x in this 
 * file always refers to Sections of RFC 3448 - LG.
 *
 *----------------------------------------------------------------------------
 *
 */
#include "config.h"
#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <sys/time.h>
#include <netinet/in.h>
#include "tv.h"
#include "tfrc_snd.h"

#define MY_MIN(A,B)     (A<B? A:B)
#define MY_MAX(A,B)	(A>B? A:B)

#define TFRC_RTT_INIT       1000000               /* intial RTT value  = 1sec */
#define TFRC_PSIZE_INIT        1470
#define TFRC_INIT_PCKTS_PER_RTT   4           /* start with 4 packets per RTT */
#define TFRC_MAGIC  	 0xbaef03b7	                     /* For debugging */
#define T_MBI		         64     /* the maximum inter-packet backoff   */
      					/* interval = 64 sec 		      */
/******************************************************************************/
/*                                                                            */
/******************************************************************************/
struct tfrc_snd {
        uint32_t         avg_rtt;            /* rtt as computed by the sender */
        uint16_t         new_rtt;          /* flag change in value of the RTT */
        uint32_t         pckt_size;             /* data_len of tfrc send data */
        uint32_t         x_send;
	uint32_t	 send_time;
        uint32_t         ipi;
	uint32_t	 tld;
	uint32_t	 x_recv,
			 x_calc;
	uint32_t	 p;
	struct timeval	 nofeedbk_timer;
	uint32_t        magic;
};


/******************************************************************************/
/*									      */
/******************************************************************************/

static void
validate_tfrc_sndr (struct tfrc_snd *tfrc)
{	
	assert(tfrc!=NULL);
        assert(tfrc->magic == TFRC_MAGIC);
}


static uint32_t
tfrc_transfer_rate(double p, uint32_t RTT, uint16_t s )
{
        double t1, t2, t3, t4, rtt, tRTO;
        uint32_t t;

        assert ((p>=0) && (p<=1));

        /* convert RTT from micro-sec to sec */
        rtt = ((double) RTT) / 1000000.0;
        tRTO = 4 * rtt;

        t1 = rtt * sqrt(2 * p / 3);
        t2 = (1 + 32 * p * p);
        t3 = 3 * sqrt(3 * p / 8);
        t4 = t1 + tRTO * t3 * p * t2;
        t = ((double) s) / t4;
        return (t);
}


/*
 * update sending rate per Sec 4.3 bullet 4.
 */
static uint32_t 
update_sending_rate ( struct tfrc_snd *tfrc, uint32_t now)
{
        double tmp = (double) tfrc->p/4294967295.0;

        /* update the sending rate */
        if (tfrc->p>0) {
                tfrc->x_calc = tfrc_transfer_rate (tmp, tfrc->avg_rtt, tfrc->pckt_size );
                tfrc->x_send = MY_MIN ( tfrc->x_calc, 2*tfrc->x_recv );
                tfrc->x_send = MY_MAX ( tfrc->x_send, tfrc->pckt_size/T_MBI);
                //printf (" - ");
        }
        else if (tfrc->p==0) {
                /* still in slow start */
                if ((tfrc->tld + tfrc->avg_rtt) <= now) {
			if (tfrc->x_recv!=0)
                        	tfrc->x_send = MY_MIN (2*tfrc->x_send, 2*tfrc->x_recv);
			else
                                /* FIX ME: this should only happen with the 1st */
                                /* feedback packet (sec 6.3) - check for it     */
				tfrc->x_send = 2*tfrc->x_send;
                        tfrc->x_send = MY_MAX ( tfrc->x_send, tfrc->pckt_size/tfrc->avg_rtt ); 
                        tfrc->tld=now;
                        //printf (" .ss. ");
                }
        } else 
		printf ("\n how did we end up with p=%d, %f", tfrc->p, tmp);

	assert (tfrc->x_send!=0);
	return (tfrc->x_send);
}


/******************************************************************************/
/*                                                                            */
/*  External API follows...						      */
/*                                                                            */
/******************************************************************************/
uint32_t
tfrc_rtt_flag (struct tfrc_snd *tfrc) 
{
	assert (tfrc!=NULL);
	return (tfrc->new_rtt==1);
}


void 
tfrc_set_rtt_flag (struct tfrc_snd *tfrc, int a)
{
	assert (tfrc!=NULL);
	tfrc->new_rtt=a;
}


uint32_t
tfrc_get_avg_rtt (struct tfrc_snd *tfrc)
{
	assert (tfrc!=NULL);
	return (tfrc->avg_rtt);
}

uint32_t
tfrc_get_ipi (struct tfrc_snd *tfrc)
{
        assert (tfrc!=NULL);
        return (tfrc->ipi);
}

void
tfrc_set_ipi (struct tfrc_snd *tfrc, uint32_t a)
{
        assert (tfrc!=NULL);
        tfrc->ipi=a;
}

void
tfrc_set_sndr_psize (struct tfrc_snd *tfrc, uint32_t a)
{
        assert (tfrc!=NULL);
        tfrc->pckt_size = a;
}

uint32_t
tfrc_get_send_rate (struct tfrc_snd *tfrc)                                        
{
        assert (tfrc!=NULL);
        return (tfrc->x_send);
}

uint32_t
tfrc_get_recv_rate (struct tfrc_snd *tfrc)
{
        assert (tfrc!=NULL);
        return (tfrc->x_recv);
}


uint32_t
tfrc_send_ok (struct tfrc_snd *tfrc, uint32_t now)
{
        assert (tfrc!=NULL);

        if (now < tfrc->send_time) {
                return (0);
        }
        else {
                tfrc->send_time+=tfrc->ipi;
                return (1);
        }
}

/*
 * Sec 4.3
 */
void         
tfrc_process_feedback (struct tfrc_snd *tfrc, struct timeval tnow,
			 uint32_t tdelay, uint32_t x_recv, uint32_t send_ts, uint32_t p) 
{
 	uint32_t rtt,
		 now,
		 prev_rtt;

	assert (tfrc!=NULL);
	

	now = timeval_to_usec (tnow);

	/* compute the RTT */
	rtt =  now - send_ts - tdelay;
	prev_rtt = tfrc->avg_rtt;
	if (tfrc->new_rtt==2) {
		/* no feedback has been received before */
		tfrc->avg_rtt = rtt;
		tfrc->new_rtt=1;
	}
	else 
		tfrc->avg_rtt = (0.9)*tfrc->avg_rtt + (1-0.9)*rtt;

	if (abs (prev_rtt - tfrc->avg_rtt) >= 10)  {
		/* set new rtt flag*/
		tfrc->new_rtt=1;
	}

	tfrc->x_recv = x_recv;
	tfrc->p = p;

	/* compute new send rate */
	tfrc->x_send = update_sending_rate (tfrc, now);
	tfrc->ipi = (tfrc->pckt_size*1000000)/tfrc->x_send;

	//printf (" %2d, %2d Mbps | %8.6f rtt:%3d, %d ipi:%d\n", 
	//			(x_recv*8)/1000000, 
	//			(tfrc->x_send*8)/1000000, 
	//			tfrc->p/4294967295.0, rtt, tfrc->avg_rtt, tfrc->ipi);

	/* reset the nofeedback timer */
        tv_add_usec (&tnow, MY_MAX (2*tfrc->avg_rtt, 2*tfrc->ipi));
        tfrc->nofeedbk_timer.tv_sec = tnow.tv_sec;
        tfrc->nofeedbk_timer.tv_usec = tnow.tv_usec;

}

/*
 * Sec 4.4
 */
void
tfrc_poll_rcvr_feedback(struct tfrc_snd *tfrc, struct timeval tnow)
{
        //uint32_t now;
        assert (tfrc!=NULL);

        if (tv_gt (tnow, tfrc->nofeedbk_timer)){

                /*if (tfrc->x_recv) {
                        if (tfrc->x_calc > 2*tfrc->x_recv || tfrc->x_calc==0)
                                tfrc->x_recv = MY_MAX ( tfrc->x_recv/2, tfrc->pckt_size/(2*T_MBI));
                        else
                                tfrc->x_recv = tfrc->x_calc/4;
                        now = timeval_to_usec (tnow);
                        tfrc->x_send = update_sending_rate (tfrc, now);
                }
                else {
                        tfrc->x_send = MY_MAX ( tfrc->x_send/2, tfrc->pckt_size/T_MBI);
                }*/
		
		/* just half the send rate, much like ns2 does - LG */
		tfrc->x_send = MY_MAX ( tfrc->x_send/2, tfrc->pckt_size/T_MBI);

                assert (tfrc->x_send!=0);

                tfrc->ipi = (tfrc->pckt_size*1000000)/tfrc->x_send;

                //printf (" ---, %2d Mbps | %8.6f ---- %d ipi:%d\n",
                //                (tfrc->x_send*8)/1000000,
                //               tfrc->p/4294967295.0, tfrc->avg_rtt, tfrc->ipi);

                /* reset the timer */
                tv_add_usec (&tnow, MY_MAX (2*tfrc->avg_rtt, 2*tfrc->ipi));
                tfrc->nofeedbk_timer.tv_sec = tnow.tv_sec;
                tfrc->nofeedbk_timer.tv_usec = tnow.tv_usec;
        }
}


struct tfrc_snd *
tfrc_sndr_init (struct timeval tnow, uint32_t size)
{
	struct tfrc_snd *tfrc;
	uint32_t  now = timeval_to_usec (tnow);

	tfrc= (struct tfrc_snd *) malloc (sizeof (struct tfrc_snd));
        tfrc->new_rtt            = 2;               
        tfrc->avg_rtt            = TFRC_RTT_INIT; 
        tfrc->x_send             = TFRC_INIT_PCKTS_PER_RTT*TFRC_PSIZE_INIT;           
        tfrc->ipi                = TFRC_RTT_INIT/TFRC_INIT_PCKTS_PER_RTT; 
	tfrc->p			 = 0;
	tfrc->x_recv             = 0;
	tfrc->x_calc             = 0;
	tfrc->pckt_size          = size;

	tfrc->tld 		 = now;
 	tfrc->send_time	 	 = now; 
	tfrc->nofeedbk_timer.tv_sec  = tnow.tv_sec;
	tfrc->nofeedbk_timer.tv_usec = tnow.tv_usec;
	tv_add (&(tfrc->nofeedbk_timer), 2*TFRC_RTT_INIT);		

	tfrc->magic = TFRC_MAGIC;
	return (tfrc);
}

