
/*
 * FILE:   controller.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.46 $
 * $Date: 2005/06/23 22:24:05 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "version.h"

#include "video_types.h"
#include "video_display.h"
#include "rtp/rtp.h"
#include "rtp/rtp_callback.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_codec.h"
#include "pdb.h"
#include "tv.h"
#include "transmit.h"
#include "tfrc.h"
#include "environment.h"
#include "controller.h"

/* Use locks */
// #define ENABLE_EXCLUSION

/* Use the experimental TX code */
//#define EXPERIMENTAL_TXSEND

typedef enum {
	t_trans,
	t_recep,
} t_subsystem;

t_ctrlr_err	ctrlr_cleanup( t_subsystem ); 

/*
 * Controller state
 */ 
static struct t_ctrlr_state {
	struct timeval  start_time; 		/* Transmission starting time */

	struct rtp		*network_device;	
	
	struct vidcap		*capture_device;
	vidcap_frame_enc_t	 capture_device_enc;

	struct display		*display_device;
	struct pdb		*participants;
	struct video_tx		*tx;
		
	int			 trans_working;
	int			 recep_working;
	
	/* Transmission and reception threads */
	pthread_t       trans_thread;
	pthread_t       recep_thread;
	
	/* Synchronization locks */
	pthread_mutex_t trans_working_lock;
	pthread_mutex_t recep_working_lock;
	pthread_mutex_t network_device_lock;
} ctrlr;

/*
 * Subsystems initialization
 */
int
ctrlr_init_video_codecs(struct run_env *env)
{
	unsigned int		 nc;
	unsigned int		 i;
	const char		*codec_name	= NULL;
	
	assert( env        != NULL );
	assert( env->codec != NULL );

	vcodec_init();
	nc = vcodec_get_num_codecs();
	for (i = 0; i < nc; i++) {
		if (vcodec_can_encode(i)) {
			printf("Video encoder : %s\n", vcodec_get_description(i));
		}
		if (vcodec_can_decode(i)) {
			printf("Video decoder : %s\n", vcodec_get_description(i));
		}

		codec_name = vcodec_get_name(i);
		
		if(codec_name != NULL) {
			/* Map static and "well-known" dynamic payload types */
			if (strcmp(codec_name, "uv_yuv422_8") == 0) {
				vcodec_map_payload(30, i);	/* 30 : Uncompressed YUV 4:2:2 8 bit */
			}

			if (strcmp(vcodec_get_name(i), "uv_yuv422_10") == 0) {
				vcodec_map_payload(32, i);      /* 32 : Uncompressed YUV 4:2:2 10 bit */
			}
			
			if (strcmp(codec_name, "dv") == 0) {
				vcodec_map_payload(123, i);	/* 123 : DV */
			}
		
		}
	}
		
	return TRUE;
}

struct display *
ctrlr_init_video_display(struct run_env* env)
{
	unsigned int             i;
	display_type_t		*dt;
	display_format_t	*format;

	assert(env->display != NULL);	
	
	ctrlr.display_device = NULL;
	
	display_init_devices(env, env->display_args);
	for (i = 0; i < display_get_device_count(); i++) {
		dt = display_get_device_details(i);

		format = dt->formats;
		
		assert(dt->name != NULL);
		
		if (strcmp(env->display, dt->name) == 0) {
			ctrlr.display_device = display_init(dt->id, format, env->display_args);
			break;
		}
	}
	display_free_devices();

	return ctrlr.display_device;
}

struct vidcap *
ctrlr_init_video_capture(struct run_env *env)
{
	unsigned int		 i;
	struct vidcap_type	*vt;
	capture_format_t	*format;
	
	assert(env->capture != NULL);
	
	ctrlr.capture_device = NULL;
	
	vidcap_init_devices(env);
	for (i = 0; i < vidcap_get_device_count(); i++) {

		vt     = vidcap_get_device_details(i);

		/* FIXME: we use the first capture format as the default... */
		format = vt->formats;
			
		assert(vt->name != NULL);
	
		if (strcmp(vt->name, env->capture) == 0) {
			ctrlr.capture_device	 = vidcap_init(vt->id, format, env->capture_args);
			ctrlr.capture_device_enc = format->encoding;
			break;
		}
	}
	
	vidcap_free_devices();

	return ctrlr.capture_device;
}

struct rtp *
ctrlr_init_network(struct run_env * env)
{
	struct rtp     *r;
	double          rtcp_bw = 5 * 1024 * 1024;	/* FIXME */

	ctrlr.participants = pdb_init();
	
	if( ctrlr.participants != NULL ) {
		r = rtp_init(env->addr, env->rx_port, env->tx_port, 255, rtcp_bw, FALSE, rtp_recv_callback, (void *)ctrlr.participants);
		if (r != NULL) {
			pdb_add(ctrlr.participants, rtp_my_ssrc(r));
			rtp_set_option(r, RTP_OPT_WEAK_VALIDATION, TRUE);
			rtp_set_sdes(r, rtp_my_ssrc(r), RTCP_SDES_TOOL, ULTRAGRID_VERSION, strlen(ULTRAGRID_VERSION));
		}
		
		ctrlr.network_device = r;
		
		return r;
	}
	
	return NULL;
}

struct video_tx *
ctrlr_init_transmit(struct run_env * env)
{
	/* Currently this is trivial. It'll get more complex once we */
	/* have multiple codecs and/or error correction.             */
	return (ctrlr.tx = tx_init(env->mtu, env->fps));
}

t_ctrlr_err
ctrlr_init_all(struct run_env * env)
{
	assert( env != NULL );
	
	if ( ctrlr_init_video_codecs(env) == FALSE)
		return ERR_CODEC;

	if ((ctrlr.display_device = ctrlr_init_video_display(env)) == NULL)
		return ERR_DISPLAY;

	if ((ctrlr.capture_device = ctrlr_init_video_capture(env)) == NULL)
		return ERR_CAPTURE;

	if ((ctrlr.network_device = ctrlr_init_network(env)) == NULL)
		return ERR_NET;

	if ((ctrlr.tx = ctrlr_init_transmit(env)) == NULL)
		return ERR_TRANS;

	return ERR_OK;
}


/*
 * Transmission and reception threads
 */
static void *
ctrlr_trans_thread_proc(void *arg)
{
	struct video_frame	*grab_frame	= NULL,
				*dec_frame	= NULL;
	struct encoded_frame	*enc_frame	= NULL;
	struct run_env		*env		= (struct run_env*) arg;
	int			 work		= FALSE;
	
	unsigned		 encoder_id	= VCODEC_INVALID_ID;	
	const char		*encoder_name	= env->codec;
	void			*encoder	= NULL;
	vcodec_enc_opt_t         encoder_flags  = 0;
	
	unsigned		 decoder_id	= VCODEC_INVALID_ID;
	const char		*decoder_name	= NULL;
	void			*decoder	= NULL;
	
	t_usec			 allowed_time;
#ifdef EXPERIMENTAL_TXSEND
	t_usec			 diff_time;
	struct timeval		 curr_time, prev_time;
#endif

	assert(ctrlr.capture_device != NULL);
	assert(ctrlr.network_device != NULL);
	assert(ctrlr.tx             != NULL); 
	assert(env                  != NULL);

	
	debug_msg("Started transmit thread\n");

	/* Set the "we are working" flag */
	ctrlr.trans_working = TRUE;
	work = ctrlr.trans_working;

	/* 
	 * Get the decoder for decoding grabbed frames, as well as the encoder
	 * for outgoing frames
	 */
	decoder_id = vcodec_get_by_enc(ctrlr.capture_device_enc);
	encoder_id = vcodec_get_by_name(env->codec);
	
	debug_msg("Codecs chain: in=%s -> out=%s\n",
		  vcodec_get_name(decoder_id), 
		  vcodec_get_name(encoder_id));

	/*
	 * If they don't match, we'll need to decode and encode again.
	 * Note: null codecs are represented with a VCODEC_INVALID_ID. For instance,
	 * a raw capture device does not need to be decoded, so its decoder_id
	 * should be -1.
	 */
	if (decoder_id == encoder_id) {
		encoder_flags = VCODEC_ALREADY_ENCODED;
	} else {
		/* If needed, create the decoder for decoding grabbed frames */
		if(decoder_id != VCODEC_INVALID_ID) {
			if (vcodec_can_decode(decoder_id) == FALSE) {
				fprintf(stderr, "Error: %s decoder can not decode grabbed frames.\n", 
					decoder_name);
				work = FALSE;	
			}			
			
			decoder    = vcodec_decoder_create (decoder_id);
			if (decoder == NULL) {
				fprintf(stderr, "Error: creating %s decoder for grabbed frames.\n", 
					decoder_name);
				ctrlr_cleanup(t_recep);
				work = FALSE;	
			}
			
			/* Create the frame for decoding grabbed images */
			dec_frame = malloc(sizeof(struct video_frame));
			if (dec_frame == NULL) {
				fprintf(stderr, "Error: allocating memory for decoding grabbed frames.\n");
				work = FALSE;	
			} else {				
				dec_frame->width       = display_get_width(vcodec_decoder_get_size(decoder_id));
				dec_frame->height      = display_get_height(vcodec_decoder_get_size(decoder_id));
				dec_frame->colour_mode = vcodec_decoder_get_cm(decoder_id);
				dec_frame->data_len    = vcodec_decoder_get_framelen(decoder_id);
				dec_frame->data        = malloc(dec_frame->data_len);
				if (dec_frame->data == NULL) {
					fprintf(stderr, "Error: allocating memory for decoding grabbed frames.\n");
					work = FALSE;	
				}
			}
		} else {
			debug_msg("Invalid decoder\n");
		}
	}

	/* Create an encoder: we will always need it */
	if (vcodec_can_encode(encoder_id) == FALSE) {
		fprintf(stderr, "Error: %s encoder can not encode outgoing frames.\n", 
			encoder_name);
		work = FALSE;	
	} else {
		encoder = vcodec_encoder_create (encoder_id, env->mtu);
		if (encoder == NULL) {
			fprintf(stderr, "Error: creating %s encoder for outgoing frames.\n", 
				encoder_name);
			work = FALSE;	
		}			

		enc_frame = (struct encoded_frame*) malloc(sizeof(struct encoded_frame));
		if (enc_frame == NULL) {
			fprintf(stderr, "Error: allocating memory for encoding outgoing frames.\n");
			work = FALSE;	
		}				
	}	

	/*
	 *  The time we give to transmit a frame
	 *  It should be less than 1/FPS, because we spend some time grabbing...
	 */
	allowed_time =  1000000.0 / env->fps;

	while (work) {
#ifdef EXPERIMENTAL_TXSEND
		gettimeofday(&prev_time, NULL);
#endif

		/* Capture, decode (if needed), encode and transmit a frame... */
		grab_frame = vidcap_grab(ctrlr.capture_device);
		if (grab_frame != NULL) {
			if (decoder != NULL) {
				vcodec_decode_raw(decoder_id, decoder, grab_frame, dec_frame);
				if(dec_frame != NULL) {
					vcodec_encode(encoder_id, encoder, dec_frame, enc_frame, encoder_flags);
				}
			} else {
				vcodec_encode(encoder_id, encoder, grab_frame, enc_frame, encoder_flags);
			}

			if (enc_frame != NULL) {
#ifdef EXPERIMENTAL_TXSEND
				/* Determine how much time has beeen spent with the */
				/* decoding/encoding process, and substract it to   */
				/* the allowed_time for sending the frame.          */
				gettimeofday(&curr_time, 0);
				diff_time = tv_diff_usec(curr_time, prev_time);

#ifdef ENABLE_EXCLUSION			
				pthread_mutex_lock(&ctrlr.network_device_lock);
				tx_send(ctrlr.tx, enc_frame, ctrlr.network_device, allowed_time - diff_time);
#endif
#else
#ifdef ENABLE_EXCLUSION
				pthread_mutex_lock(&ctrlr.network_device_lock);
#endif
				tx_send(ctrlr.tx, enc_frame, ctrlr.network_device, allowed_time);
#endif
#ifdef ENABLE_EXCLUSION
				pthread_mutex_unlock(&ctrlr.network_device_lock);
#endif
				free(grab_frame);
			}
		}

		/* Check if this thread must work... */
		pthread_mutex_lock(&ctrlr.trans_working_lock);
		work = ctrlr.trans_working;
		pthread_mutex_unlock(&ctrlr.trans_working_lock);
	};

	/* Free memory of temporal frames */	
	if (encoder != NULL) {
		vcodec_encoder_destroy (encoder_id, encoder);
		free(enc_frame);
	}

	if (decoder != NULL) {
		vcodec_decoder_destroy (decoder_id, decoder);

		if(dec_frame->data != NULL)
			free(dec_frame->data);
		
		free(dec_frame);
	}
		
	ctrlr.trans_working = FALSE;
	
	ctrlr_cleanup(t_trans);
	
	return NULL;
}

static void *
ctrlr_recep_thread_proc(void *arg)
{
	struct timeval	 timeout;
	struct timeval	 curr_time;
	struct pdb_e	*cp;
	uint32_t	 ts   = 0;
	struct run_env	*env  = (struct run_env*) arg;	
	int		 work;
	display_result_t display_result = DISPLAY_OK;

	assert( ctrlr.participants   != NULL );
	assert( ctrlr.network_device != NULL );

	debug_msg("Started receiver thread\n");
		
	ctrlr.recep_working = TRUE;
	work = ctrlr.recep_working;

	timeout.tv_sec  = 0;
	timeout.tv_usec = 999999 / env->fps;
	
	gettimeofday(&(ctrlr.start_time), NULL);
	while (work) {
		/* Housekeeping and RTCP... */
		gettimeofday(&curr_time, NULL);
		ts = tv_diff(curr_time, ctrlr.start_time) * 90000;

#ifdef ENABLE_EXCLUSION
		pthread_mutex_lock(&ctrlr.network_device_lock);
#endif
		rtp_update(ctrlr.network_device, curr_time);
		rtp_send_ctrl(ctrlr.network_device, ts, 0, curr_time);		

#ifdef ENABLE_EXCLUSION
		pthread_mutex_unlock(&ctrlr.network_device_lock);
#endif
		
		/*
		 * Receive packets from the network...
		 */
		rtp_recv(ctrlr.network_device, &timeout, ts);

		/*
		 * Decode and render for each participant in the
		 * conference...
		 */
		cp = pdb_iter_init(ctrlr.participants);
		while (cp != NULL && display_result == DISPLAY_OK) {
			/* Decode and render video... */
			gettimeofday(&curr_time, NULL);
			display_result = pbuf_decode(cp->playout_buffer, curr_time,
						     ctrlr.display_device);
			pbuf_remove(cp->playout_buffer, curr_time);
			cp = pdb_iter_next(ctrlr.participants);
		}
		pdb_iter_done(ctrlr.participants);

		if( display_result != DISPLAY_OK )
			ctrlr_done();

		/* Use lock trans_lock and read everything_is_done */
		pthread_mutex_lock(&ctrlr.recep_working_lock);		
		work = ctrlr.recep_working;
		pthread_mutex_unlock(&ctrlr.recep_working_lock);
	}

	debug_msg("Stopped receiver thread\n");

	ctrlr.recep_working = FALSE;
	
	ctrlr_cleanup(t_recep);
	
	return NULL;
}


/*
 * Starts the transmission
 */
t_ctrlr_err
ctrlr_start(struct run_env* env)
{
	ctrlr.trans_working = FALSE;
	ctrlr.recep_working = FALSE;
		
	/*
	 * Creates a transmission and reception thread
	 */ 
	if( pthread_mutex_init(&ctrlr.trans_working_lock, NULL) != 0 ) {
		perror("Creating the transmit mutex:");
		return ERR_THREADS;
	}
	
	if( pthread_mutex_init(&ctrlr.recep_working_lock, NULL) != 0 ) {
		perror("Creating the reception mutex:");
		return ERR_THREADS;
	}	
	
	if( pthread_mutex_init(&ctrlr.network_device_lock, NULL) != 0 ) {
		perror("Creating the network device mutex:");
		return ERR_THREADS;
	}

	if( pthread_create(&ctrlr.recep_thread, NULL, ctrlr_recep_thread_proc, env) != 0 ) {
		perror("Creating the reception thread:");
		return ERR_THREADS;
	}

	/* Wait for the reception thread to be working... */
	//while (ctrlr.recep_working != TRUE) {};	
	
	if( pthread_create(&ctrlr.trans_thread, NULL, ctrlr_trans_thread_proc, env) != 0 ) {
		perror("Creating the transmission thread:");
		return ERR_THREADS;
	}

	return ERR_OK;
}


/*
 * Stop a subsystem.
 * Called by external sources or by subsystems to make a cleanup when
 * they stop.
 */
t_ctrlr_err
ctrlr_cleanup( t_subsystem subsys ) {
	int		 something_done = FALSE;
	uint32_t         ssrc;
	struct timeval	 curr_time;
	struct pdb_e	*cp;
	
	if( subsys == t_trans && ctrlr.trans_working ) {
		ctrlr.trans_working = FALSE;

		something_done = TRUE;
		
		if( pthread_self() != ctrlr.trans_thread ) {
			debug_msg("Waiting transmit thread...\n");
			pthread_join(ctrlr.trans_thread, NULL);
			debug_msg("Transmit thread joint.\n");
			
			pthread_mutex_destroy(&ctrlr.trans_working_lock);
		}
		
	} else if( subsys == t_recep && ctrlr.recep_working) {
		ctrlr.recep_working = FALSE;

		something_done = TRUE;
		
		if( pthread_self() != ctrlr.recep_thread ) {		
			debug_msg("Waiting receiver thread...\n");
			pthread_join(ctrlr.recep_thread, NULL);
			debug_msg("Receiver thread joint!\n");
			
			pthread_mutex_destroy(&ctrlr.recep_working_lock);
		}
	}
		
	/*
	 * If both subsystems have stopped, we can make a cleanup.
	 */
	if (!ctrlr.trans_working && !ctrlr.recep_working && something_done) {
		/* Finish all the subsystems */
		debug_msg("Stopping all subsystems...\n");
		
		tx_done(ctrlr.tx);
		rtp_done(ctrlr.network_device);
		vidcap_done(ctrlr.capture_device);
		display_done(ctrlr.display_device);
		vcodec_done();
		
		/*
		 * Pause to allow playout time for any queued data to pass,
		 * then empty the playout buffers...
		 */ 
		sleep(1);
		gettimeofday(&curr_time, NULL);
		
		cp = pdb_iter_init(ctrlr.participants);
		while (cp != NULL) {
			pbuf_remove(cp->playout_buffer, curr_time);
			pbuf_done(cp->playout_buffer);
			ssrc = cp->ssrc;
			cp = pdb_iter_next(ctrlr.participants);
			pdb_remove(ctrlr.participants, ssrc, NULL);
		}
		pdb_iter_done(ctrlr.participants);
		pdb_destroy(&ctrlr.participants);
	}	

	return ERR_OK;
}


/*
 * Stop the transmission
 */
t_ctrlr_err
ctrlr_done(void)
{
	ctrlr_cleanup(t_trans);
	ctrlr_cleanup(t_recep);
	
	return ERR_OK;
}

int
ctrlr_working(void)
{
	return (ctrlr.trans_working || ctrlr.recep_working);
}
