/*
 * FILE:    main.c
 * AUTHORS: Colin Perkins <csp@csperkins.org>
 *          Ladan Gharai  <ladan@isi.edu>
 *          Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2001-2004 University of Southern California
 * Copyright (c) 2003-2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.110 $
 * $Date: 2005/05/16 10:29:04 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "version.h"
#include "environment.h"
#include "controller.h"
#include "ui/ui.h"

#ifdef HAVE_LIBSDL
#include <SDL.h>
#endif

static void
signal_handler(int signal)
{
	debug_msg("Caught signal %d\n", signal);
	if (signal == SIGINT) {
		ctrlr_done(); 	/* Signal the controller to finish */
		ui_done(); 	/* Stop the UI */
		return;
	}
	abort();
}

uint32_t  RTT = 0;

/*
 * This is the main routine for non-MacOS X platforms. If running on
 * MacOS X with Cocoa, we are an NSApplication and this function is 
 * not used.
 */
int 
main(int argc, char *argv[])
{
        struct run_env	*environ;
	int		 err_code	= ERR_OK;

        /*
         * Loads the defaults and then parses the command line
         */
        if ( (environ = env_init()) == NULL) {
                return ERR_USAGE;
	}
        if ( (err_code = env_load_cmdline(&argc, argv, environ)) != ERR_OK) {
                return err_code;
        }

#ifdef DEBUG
	printf("%s *** DEBUG BUILD - PERFORMANCE WILL SUFFER ***\n", ULTRAGRID_VERSION);
#else
	printf("%s\n", ULTRAGRID_VERSION);
#endif
	printf("Display device: %s\n", environ->display);
	printf("Capture device: %s\n", environ->capture);
	printf("Frame rate    : %d\n", environ->fps);
	printf("MTU           : %d\n", environ->mtu);
        printf("RX port       : %d\n", environ->rx_port);
        printf("TX port       : %d\n", environ->tx_port);
        printf("TX codec      : %s\n", environ->codec);
	
	/* Initialize the UI */
	if (ui_init(argc, argv, environ->display) != 0) {
		return ERR_UI;
	}

	signal(SIGINT, signal_handler);

	/* Initialize all subsystems */
	if ((err_code = ctrlr_init_all(environ)) != ERR_OK) {
		switch( err_code ) {
			case ERR_CODEC:
				fprintf(stderr,"Codec initialization error.\n");
				break;
			case ERR_DISPLAY:
				fprintf(stderr,"Display initialization error.\n");
				break;
			case ERR_CAPTURE:
				fprintf(stderr,"Capture initialization error.\n");
				break;
			case ERR_NET:
				fprintf(stderr,"Network initialization error.\n");
				break;
			case ERR_TRANS:
				fprintf(stderr,"Transmission initialization error.\n");
				break;
		}
		fprintf(stderr,"%s\n", strerror(errno));
		return err_code;	
	}
	
	/* Start transmission */
	if ((err_code = ctrlr_start(environ)) != ERR_OK)
		return err_code;

	/* 
	 * Wait for the recep/trans threads initialization and, then, wait for
	 * their dead...
	 */ 
	while (!ctrlr_working())	sleep(1);
	while (ctrlr_working())		sleep(1);
	        
	return ERR_OK;
}

