/*
 * FILE:   trace.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.8 $
 * $Date: 2005/04/26 14:45:05 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "trace.h"


#define MAX_TRACE_EVENT_T	255	/* max number of trace events */

/* Table entry */
struct trace_table_entry_t {
	trace_output_t	output_method;
	FILE*		fd;
	trace_fun_t	fun;
	char*		desc;
	struct timeval	last_event;
	
};

static struct trace_table_entry_t	trace_table[MAX_TRACE_EVENT_T];


/*
 * Format and dump (if necessary) the last event(s).
 * 
 * TODO: This function should store all the events corresponding to the same
 * moment in time and, when a new time is detected, dump all old data with
 * the appropiate format.
 */ 
static void
record_event( struct timeval* ts, trace_event_t ev, trace_value_t val) {
	UNUSED(ts);
	UNUSED(val);
	UNUSED(ev);
	
	//fprintf(state_trace.fd,"%s=%F\n", t_event_str[ev], val);
}

/*
 * Public API
 */
void
trace_init(trace_output_t out, void* def)
{
	unsigned int	i;
	FILE*		default_fd;

	/* Initialize the table with default values */
	if(out == t_file) {
		
		/* Open the default log file */
		default_fd = fopen(def, "w+");
		if (default_fd == NULL) {
			fprintf(stderr, "Error creating default trace file: %s.\n", strerror(errno));
			return;
		}
		
		for(i=0; i<MAX_TRACE_EVENT_T; i++) {
			trace_table[i].output_method = t_file;
			trace_table[i].fd            = default_fd;
			trace_table[i].desc          = "";
		}
		
	} else if(out == t_function) {
		
		/* Send all trace info  to the same function */
		for(i=0; i<MAX_TRACE_EVENT_T; i++) {
			trace_table[i].output_method	= t_function;
			trace_table[i].fun		= (trace_fun_t)def;
			trace_table[i].desc		= "";
		}
	}

}

/*
 * Register an event
 */
void
trace_reg(trace_event_t ev, char* desc, void* target, trace_output_t out)
{
	unsigned int index	= (unsigned int) ev;
	
	assert(index  <  MAX_TRACE_EVENT_T);
	assert(target != NULL);
	
	if( out == t_file ) {
		trace_table[index].fd = fopen((char*)target, "w" );
		
		if(trace_table[index].fd == NULL) {
			
			fprintf(stderr, "Error: opening trace file: %s\n", strerror(errno));
			
			trace_table[index].output_method	= t_none;
			trace_table[index].fd			= 0;
			
		} else {
			trace_table[index].output_method = t_file;
			trace_table[index].desc		= desc;

		}
	} else if ( out == t_function ) {
		trace_table[index].output_method	= t_function;
		trace_table[index].fun			= (trace_fun_t)target;
		trace_table[index].desc			= desc;
	}
}




void
trace_event( struct timeval* ts, trace_event_t ev, trace_value_t val)
{
	unsigned int index	= (unsigned int) ev;

	assert(index < MAX_TRACE_EVENT_T);	
	
	if( trace_table[index].output_method == t_file && trace_table[index].fd != NULL )
		record_event(ts, ev, val);
	else if( trace_table[index].output_method == t_function && trace_table[index].fun != NULL )
		(trace_table[index].fun)(ts, ev, val);
}

void
trace_done(void)
{
	unsigned int i;
	
	for(i=0; i<MAX_TRACE_EVENT_T; i++) {
		if (trace_table[i].output_method == t_file)
			fclose(trace_table[i].fd);
	}
}
