/*
 * FILE:    transmit.c
 * AUTHORS: Colin Perkins <csp@csperkins.org>
 *          Ladan Gharai  <ladan@isi.edu>
 *          Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c)      2005 University of Glasgow
 * Copyright (c) 2001-2004 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.39 $
 * $Date: 2005/08/11 19:08:02 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "video_types.h"
#include "video_display.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"	/* For struct coded_data */
#include "tv.h"
#include "transmit.h"
#include "video_capture.h"
#include "video_codec.h"

#define TRANSMIT_MAGIC	0xe80ab15f

#ifdef EXPERIMENTAL_TXSEND
/* We don't want to send more than these packets together... */
#define MAX_PACKETS_BURST	8
#endif

struct video_tx {
	uint32_t	 magic;
	unsigned	 mtu;
	unsigned	 fps;
};

struct video_tx *
tx_init(unsigned mtu, unsigned fps)
{
	struct video_tx	*tx;

	tx = (struct video_tx *) malloc(sizeof(struct video_tx));
	if (tx != NULL) {
		tx->magic = TRANSMIT_MAGIC;
		tx->mtu   = mtu;
		tx->fps   = fps;
	}
	return tx;
}

void
tx_done(struct video_tx *tx)
{
	assert(tx->magic == TRANSMIT_MAGIC);
	free(tx);
}

/*
 * Spacing and transmission of packets.
 *
 * See RFC 3448, "4.6. Scheduling of Packet Transmissions" for some information
 * on how this should be done.
 *
 * The sending system should space packets for max(allowed_time/num_pcakets, t_gran)
 * seconds, where t_gran is the time granularity of the system (10ms on Unix).
 * After each packet is sent, the mean bitrate should be calculated and another
 * packet could be sent if we are being "late" in the sending rate.
 */
void
tx_send(struct video_tx *tx, struct encoded_frame *frame, 
	struct rtp *rtp_session, t_usec allowed_time)
{
#undef EXPERIMENTAL_TXSEND
#ifdef EXPERIMENTAL_TXSEND
	struct encoded_data	*node			= NULL;
	int			 pos			= 0;
	static t_usec		 ts			= 0;
	unsigned int		 data_real		= 0,
				 data_theo		= 0,
				 data_per_usec		= 0,
				 worst_nodes_burst	= 0;
	int			 data_gap		= 0;

	struct timeval		 time_start,
				 time_curr,
				 time_next;

	static t_usec		 time_elapsed		= 0,
				 time_per_node		= 0;

	static t_sec		 time_delta		= 0;
	static t_usec		 time_delta_usec	= 0;
		
	assert(tx             != NULL);
	assert(tx->magic      == TRANSMIT_MAGIC);
	assert(allowed_time   >  0);
	assert(frame->len     >  0);
	assert(frame->datalen >  0);

	/* Sets the next transmission time */
	gettimeofday(&time_start, NULL);
	gettimeofday(&time_next, NULL);

	data_per_usec = frame->datalen / allowed_time;
	assert(data_per_usec > 0);

	/*
	 * Give allowed_time/len usecs per node.
	 *
	 * Note: this can be unfair. It only considers the number of
	 * nodes, not their lenght! If we have 1 second, and 2 nodes, it will
	 * give 0.5 seconds per node, even if the first is 5 bytes long and
	 * the second 1500. However, we suppose that the frame will be
	 * packetized in packets of roughly same size...
	 */
	time_per_node = allowed_time / frame->len;

	/*
	 * The time_delta will set how much we will sleep each time (in secs).
	 */
	time_delta_usec = max(time_per_node/2, TIME_GRANULARITY/2);
	time_delta = time_delta_usec / 1000000.0;

	assert(time_delta > 0);

	/*
	 * Calculate what is number of packets sent together in the worst
	 * case, due to the time granularity.
	 * The highest error for a sleep call is TIME_GRANULARITY*2, and 
	 * then we should send:
	 */
	worst_nodes_burst = (TIME_GRANULARITY*2)/time_per_node;

	ts = get_local_mediatime();

	/* Go through the list of nodes... */
	node = vcodec_encode_get(frame, pos);
	while(node != NULL) {
		gettimeofday(&time_curr, NULL);
		time_elapsed = tv_diff_usec(time_curr, time_start);

		/* Calculate how late or soon we are.                           */
		/* Now, data_theo corresponds to the time_allowed provided,     */
		/* which is fairly divided between all the packets, but it      */
		/* should be determined by the congestion control algorithm...  */
		data_theo = data_per_usec * time_elapsed;
		data_gap  = data_theo     - data_real;

		/* If we are late sending data (or late in time), send a packet */
		if ((data_gap >= 0) || (time_elapsed > allowed_time)) {
			rtp_send_data_hdr(rtp_session, ts,
					  vcodec_get_payload(frame->id),
					  node->mark, 0, 0,
					  (char *) node->header, node->header_len,
					  node->data, node->data_len, 0, 0, 0);
			data_real += node->header_len + node->data_len;
			node = vcodec_encode_get(frame, ++pos);

			/* Next time when we will check if we can send a packet */
			tv_add(&time_next, time_delta);			
		} else {
			/* If a sleep error could force us to send too many */
			/* packets together, we need to do a busy-wait...   */
			if(worst_nodes_burst > MAX_PACKETS_BURST) {
				while (tv_gt(time_next, time_curr)) {
					gettimeofday(&time_curr, NULL);
				}
			} else {
				t_usec		delay;
				struct timespec	ts;			

				if( tv_gt(time_next, time_curr) ) {
					delay = tv_diff_usec(time_next, time_curr);
					ts.tv_sec  = delay / 1000000;
					ts.tv_nsec = (delay % 1000000) * 1000;
					nanosleep(&ts, NULL);		
				}
			}
		}
	}
#else
	struct encoded_data	*node	= NULL;
	int			 pos	= 0;
	static t_usec		 ts	= 0;
	static t_sec		 time_per_node;	/* How much time we give per node */
	struct timeval		 next_trans_time;

	assert(tx           != NULL);
	assert(tx->magic    == TRANSMIT_MAGIC);
	assert(allowed_time >  0);
	assert(frame->len   >  0);
	
	/* Sets the next transmission time */
	gettimeofday(&next_trans_time, NULL);

	/*
	 * Give allowed_time/len secs. per node.
	 * NOTE: this can be really unfair! It only considers the number of
	 * nodes, not their length! If we have 1 second, and 2 nodes, it will
	 * give 0.5 seconds per node, even if the first is 5 bytes long and
	 * the second 1500...
	 */
	time_per_node = (allowed_time / frame->len) / 1000000.0;
	assert(time_per_node > 0);

	/* tv_add expects seconds */
	tv_add(&next_trans_time, time_per_node );

	ts = get_local_mediatime();
	
	/* Go through the list of nodes... */
	node = vcodec_encode_get(frame, pos);
	while(node != NULL) {
		/* Sleep until it is time to transmit.           */
		/* If the system is not fast enough, do nothing. */ 
		sleep_until(next_trans_time);
		rtp_send_data_hdr(rtp_session, ts,
				  vcodec_get_payload(frame->id),
				  node->mark, 0, 0,
				  (unsigned char *) node->header, node->header_len,
				  node->data, node->data_len, 0, 0, 0);
		
		/* Calculate next transmission time */
		tv_add(&next_trans_time, time_per_node);
		node = vcodec_encode_get(frame, ++pos);
	}
#endif
}

