/*
 * FILE:   UGRunEnv.m
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.1 $
 * $Date: 2005/06/08 12:07:46 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "trace.h"
#include "environment.h"
#include "controller.h"
#include "video_types.h"
#include "video_capture.h"
#include "video_display.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_codec.h"

#import <Cocoa/Cocoa.h>

#import "ui/cocoa/UGRunEnv.h"

@implementation UGRunEnv

/*******************************************************************************
* Constructors and destructors
******************************************************************************/
- (UGRunEnv*)init
{
	self = [super init];
	
	if ( self ) {
		env = env_init();
		
		captureDevices = NULL;
		displayDevices = NULL;
		videoCodecs    = NULL;		
	}
	
	return self;
}

- (void)dealloc
{
	if (env != NULL)
		env_done(env);

	[super dealloc];
}

/*******************************************************************************
* Lists of display devices, capture devices or codecs
******************************************************************************/

- (void)initVideoCodecs
{
	unsigned int		 i;
	unsigned int		 num_codecs;
	NSString		*name,
				*desc;
	
	vcodec_init();
	num_codecs = vcodec_get_num_codecs();
	
	videoCodecs = [NSMutableDictionary dictionaryWithCapacity:num_codecs];
	
	for(i=0; i < num_codecs; i++) {
		name = [NSString stringWithCString:vcodec_get_name(i)];
		desc = [NSString stringWithCString:vcodec_get_description(i)];
		[videoCodecs setObject:name forKey:desc];
	}
	
	[videoCodecs retain];	
}

- (void)initDisplayDevices
{
	display_type_t		*dt;
	unsigned int             i;
	unsigned int		 num_displays;
	NSString		*name,
				*desc;	
	
	display_init_devices(NULL);
	num_displays = display_get_device_count();
	
	displayDevices = [NSMutableDictionary dictionaryWithCapacity:num_displays];
	
	for (i = 0; i < num_displays; i++) {
		dt = display_get_device_details(i);
		
		if(dt->name != NULL && dt->description != NULL) {		
			name = [NSString stringWithCString:dt->name];
			desc = [NSString stringWithCString:dt->description];
			[displayDevices setObject:name forKey:desc];
		}
	}
	display_free_devices();
	
	[displayDevices retain];
}


- (void)initCaptureDevices
{
	unsigned int		 i;
	unsigned int		 num_capture;
	struct vidcap_type	*vt;
	NSValue			*vt_objc;
	NSString		*desc;
	
	vidcap_init_devices(env);
	
	num_capture = vidcap_get_device_count();
	
	captureDevices = [NSMutableDictionary dictionaryWithCapacity:num_capture];
	
	/* Store the list of capture devices in a MutableDictionary */ 
	for (i = 0; i < num_capture; i++) {
		vt = vidcap_get_device_details(i);
		
		if(vt && vt->name != NULL && vt->description != NULL) {
			desc    = [NSString stringWithCString:vt->description];
			vt_objc = [NSValue value:vt withObjCType:@encode(struct vidcap_type)];
			[vt_objc retain];
			[captureDevices setObject:vt_objc forKey:desc];
		}
	}
	
	vidcap_free_devices();

	[captureDevices retain];
}


/*******************************************************************************
* Data model glue
******************************************************************************/
- (NSString*)CaptureDevice
{
	if (env->capture != NULL)
		return [NSString stringWithCString:env->capture];
	else
		return [NSString string];
}

- (void)setCaptureDevice:(NSString*)arg
{
	env->capture = [arg cString];
}

- (NSString*)CaptureDeviceArg
{
	if (env->capture_args != NULL)
		return [NSString stringWithCString:env->capture_args];
	else
		return [NSString string];
}

- (void)setCaptureDeviceArg:(NSString*)arg
{
	if(env->capture_args != NULL)
		free(env->capture_args);
	
	env->capture_args = malloc(sizeof(char)*[arg length]);
	[arg getCString:env->capture_args];
}

- (NSString*)CaptureDeviceDesc
{
	NSArray			*devices;
	struct vidcap_type	 vt;
	NSValue			*vt_objc;	
	unsigned int		 i;
	
	if (env->capture != NULL) {
		devices = [captureDevices allValues];
	
		for ( i=0; i<[devices count]; i++) {
			vt_objc = (NSValue*)[devices objectAtIndex:i];
			[vt_objc getValue:&vt];
			
			if (vt.name != NULL) {
				if (strcmp(vt.name, env->capture) == 0) {					
					return [NSString stringWithCString:vt.description];
				}
			}
		}
	}
	
	return nil;
}

- (void)setCaptureDeviceDesc:(NSString*)arg
{
	const char	*name = [self getCaptureDeviceNameByDesc:arg];
	
	if (name != nil) {
		env->capture = name;
	} else {
		debug_msg("Capture device not found for selected description.\n");
	}
}

- (unsigned int)CaptureDeviceXSize
{
	return env->capture_width;
}

- (void)setCaptureDeviceXSize:(unsigned int)x
{
	env->capture_width = x;
}

- (unsigned int)CaptureDeviceYSize
{
	return env->capture_height;
}

- (void)setCaptureDeviceYSize:(unsigned int)y
{
	env->capture_height = y;
}

- (BOOL)CaptureDeviceNeedsFilename
{
	NSArray			*devices;
	struct vidcap_type	 vt;
	NSValue			*vt_objc;	
	unsigned int		 i;
	BOOL			 ret;
	
	ret = NO;
	
	debug_msg("asking if needs ile\n");
	
	if (env->capture != NULL) {
		devices = [captureDevices allValues];
		
		for ( i=0; i<[devices count]; i++) {
			vt_objc = (NSValue*)[devices objectAtIndex:i];
			[vt_objc getValue:&vt];
			
			if (vt.name != NULL) {
				if (strcmp(vt.name, env->capture) == 0) {
					ret = (vt.props & VIDCAP_PARAM_FILENAME ? YES : NO);
					
					debug_msg("Name=%s, Props=%d\n", vt.name, vt.props);
				}
			}
		}
	}
	
	return ret;	
}

- (NSString*)DisplayDevice
{
	if (env->display != NULL)
		return [NSString stringWithCString:env->display];
	else
		return [NSString string];
}

- (void)setDisplayDevice:(NSString*)arg
{
	env->display = [arg cString];
}

- (NSString*)DisplayDeviceArg
{
	if (env->display_args != NULL)
		return [NSString stringWithCString:env->display_args];
	else
		return [NSString string];
}

- (void)setDisplayDeviceArg:(NSString*)arg
{
	env->display_args = [arg cString];
}

- (NSString*)DisplayDeviceDesc
{
	NSString	*desc;
	
	desc = [NSString stringWithCString:env->display];
	return [[displayDevices allKeysForObject:desc] objectAtIndex:0];
}

- (void)setDisplayDeviceDesc:(NSString*)arg
{
	env->display = [self getDisplayDeviceNameByDesc:arg];
}

- (unsigned int)FPS
{
	return env->fps;
}

- (void)setFPS:(unsigned int)f
{
	env->fps = f;
}

- (NSString*)HostName
{
	if (env->addr != NULL) {
		return [NSString stringWithCString:env->addr];
	} else {
		return [NSString string];
	}
}

- (void)setHostName:(NSString*)arg
{
	if(env->addr != NULL)
		free(env->addr);
	
	env->addr = malloc(sizeof(char)*[arg length]);
	[arg getCString:env->addr];
}

- (unsigned int)MTU
{
	return env->mtu;
}

- (void)setMTU:(unsigned int)mtu
{
	env->mtu = mtu;
}

- (unsigned int)RXPort
{
	return env->rx_port;
}

- (void)setRXPort:(unsigned int)port
{
	//env->rx_port = (port & ~1);
	env->rx_port = port;
}

- (unsigned int)TXPort
{
	return env->tx_port;
}

- (void)setTXPort:(unsigned int)port
{
	//env->tx_port = (port & ~1);
	env->tx_port = port;
}

- (NSString*)VideoCodec
{
	if (env->codec != NULL)
		return [NSString stringWithCString:env->codec];
	else
		return [NSString string];
}

- (void)setVideoCodec:(NSString*)arg
{
	env->codec = [arg cString];
}

- (NSString*)VideoCodecDesc
{
	NSString	*desc;
	
	desc = [NSString stringWithCString:env->codec];
	return [[videoCodecs allKeysForObject:desc] objectAtIndex:0];
}

- (void)setVideoCodecDesc:(NSString*)arg
{
	env->codec = [self getVideoCodecNameByDesc:arg];
}

- (struct run_env*)getRunEnv
{
	struct run_env	*lenv;
	
	lenv = malloc(sizeof(struct run_env));
	if(lenv != NULL) {
		lenv->addr           = [[self HostName] cString];
		lenv->fps            =  [self FPS];
		lenv->rx_port        =  [self RXPort];
		lenv->tx_port        =  [self TXPort];
		lenv->mtu            =  [self MTU];
		lenv->codec          = [[self VideoCodec] cString];
		lenv->capture        = [[self CaptureDevice] cString];
		lenv->capture_args   = [[self CaptureDeviceArg] cString];
		lenv->capture_height =  [self CaptureDeviceYSize];
		lenv->capture_width  =  [self CaptureDeviceXSize];
		lenv->display        = [[self DisplayDevice] cString];
		lenv->display_args   = [[self DisplayDeviceArg] cString];        
	}
	
	return lenv;
}

/*******************************************************************************
* Data model glue
******************************************************************************/
- (unsigned int)countOfCaptureDevices
{
	if (captureDevices == NULL) {
		[self initCaptureDevices];
	}
	
	if (captureDevices != NULL) {
		return [captureDevices count];
	} else {
		return 0;
	}
}

- (NSString*)objectInCaptureDevicesAtIndex:(unsigned int)index
{
	if (index <= [self countOfCaptureDevices]) {
		return [[captureDevices allKeys] objectAtIndex:index];
	} else {
		return nil;
	}
}

- (const char*)getCaptureDeviceNameByDesc:(NSString*)desc
{
	struct vidcap_type	 vt;

	if ([self countOfCaptureDevices] > 0) {
		[[captureDevices valueForKey:desc] getValue:&vt];
		return vt.name;
	} else {
		return nil;
	}
}

- (unsigned int)countOfDisplayDevices
{
	if (displayDevices == NULL) {
		[self initDisplayDevices];
	}
	
	if (displayDevices != NULL) {
		return [displayDevices count];
	} else {
		return 0;
	}
}

- (NSString*)objectInDisplayDevicesAtIndex:(unsigned int)index
{
	if (index <= [self countOfDisplayDevices]) {
		return [[displayDevices allKeys] objectAtIndex:index];
	} else {
		return nil;
	}
}

- (const char*)getDisplayDeviceNameByDesc:(NSString*)desc
{
	if ([self countOfDisplayDevices] > 0) {
		return [[displayDevices valueForKey:desc] cString];
	} else {
		return nil;
	}
}


- (unsigned int)countOfVideoCodecs
{
	if (videoCodecs == NULL)
		[self initVideoCodecs];
	
	if (videoCodecs != NULL) {
		return [videoCodecs count];
	} else {
		return 0;
	}
}

- (NSString*)objectInVideoCodecsAtIndex:(unsigned int)index
{
	if (index <= [self countOfVideoCodecs]) {
		return [[videoCodecs allKeys] objectAtIndex:index];
	} else {
		return nil;
	}
}

- (const char*)getVideoCodecNameByDesc:(NSString*)desc
{
	if ([self countOfVideoCodecs] > 0) {
		return [[videoCodecs valueForKey:desc] cString];
	} else {
		return nil;
	}
}

@end
