/*
 * FILE:   firewire_dv_linux.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *    This product includes software developed by the University of Glasgow.
 * 
 * 4. The name of the University may not be used to endorse or promote 
 *    products derived from this software without specific prior written 
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.11 $
 * $Date: 2005/06/09 13:33:31 $
 *
 */

#include "config.h"
#include "config_unix.h"

#ifdef HAVE_FIREWIRE_DV_LINUX

#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_capture.h"
#include "video_display.h"
#include "video_capture/firewire_dv_linux.h"

#include <dv1394.h>
#include <sys/poll.h>

#ifndef DV1394_IOC_INIT
#define DV1394_IOC_INIT DV1394_INIT
#endif /* DV1394_IOC_INIT */

#define RING_FRAMES     20
#define DEFAULT_CHANNEL 63
#define NUM_FORMATS     2

/*
 * Linux DV implementation
 * 
 * This is an experimental DV implementation. There is no multi-threaded grabbing,
 * and the format is hardcoded for PAL.
 */

/*
 * DV internal state
 */
struct vidcap_dvlinux_state {
	int		 fd;			/* File descriptor for the device */
	const char	*filename;		/* The corresponding file name */

	unsigned	 fps;			/* FPS */
	unsigned	 channel;
	display_size_t	 format;		/* Capture format [NTSC|PAL] */
	unsigned	 frame_len;		/* How much data should we read? */
	
	unsigned char	*ring;
	unsigned int	 ring_current;
	unsigned int	 ring_captured;
	unsigned int	 ring_done;
};

/*
 * List of possible DV device names
 */
static char* dvlinux_devices[] = {
	"/dev/dv1394",
	"/dev/dv1394-0",
	"/dev/dv1394-1",
	"/dev/dv1394-2",
	"/dev/dv1394-3",
	"/dev/dv1394/0",
	"/dev/dv1394/1",
	"/dev/dv1394/2",
	"/dev/dv1394/3",
	NULL
};

/*
 * Autodetect and open a DV device, setting the vidcap_dvlinux_state
 * with the corresponding values. 
 */
static void
dvlinux_open_default(struct vidcap_dvlinux_state* s)
{
	int	i = 0;

	/*
	 * Try to autodetect the device looping over the devices list
	 */
	while ((s->filename = dvlinux_devices[i++]) != NULL) {
		debug_msg("Probing %s...\n", s->filename);
		if ((s->fd = open( s->filename, O_RDONLY )) != -1) {
			return;
		}
	}
		
	s->fd = -1;	
}

/*
 * Initialize the internal state from the environment data.
 */
static struct vidcap_dvlinux_state *
vidcap_dvlinux_state_init (capture_format_t *format, const char *args)
{
	
	struct vidcap_dvlinux_state *s = NULL;
	
	s = (struct vidcap_dvlinux_state*) malloc(sizeof(struct vidcap_dvlinux_state));
	if (s != NULL) {
		s->fd  = -1;

		/* capture_args contains the name of the device/filename */
		s->filename = args;
		s->fps      = format->fps;
		s->format   = format->size;
		s->channel  = DEFAULT_CHANNEL;

		/*
		 * In Linux, we must perform reads of 144000 bytes for PAL
		 * and 120000 for NTSC
		 */		
		if (s->format == DS_PAL) {
			s->frame_len = 144000;
		} else if (s->format == DS_NTSC) {
			s->frame_len = 120000;
		}
	}

	return s;
}

void *
vidcap_dvlinux_init(capture_format_t *format, const char *args)
{
	struct vidcap_dvlinux_state	*s;
	struct dv1394_init		 init_params;

	debug_msg("Initializing linux DV driver...\n");
	
	/*
	 * Get the state from the environment and open the device
	 */
	s = vidcap_dvlinux_state_init( format, args );
	if( s->filename != NULL )
		s->fd  = open(s->filename, O_RDONLY);
	else
		dvlinux_open_default( s );
	
	if (s->fd < 0) {
		fprintf(stderr, "Error: Unable to open DV device/file %s: %s",
			s->filename, strerror(errno));
		free(s);
		return NULL;
	} else {
		debug_msg("Using DV device: %s\n", s->filename);
	}

	/*
	 * Configure the capture device.
	 */
	init_params.api_version = DV1394_API_VERSION;
	init_params.channel     = s->channel;
	init_params.n_frames    = RING_FRAMES;
	init_params.cip_n       = 0;
	init_params.cip_d       = 0;
	init_params.syt_offset  = 0;
	
	if (s->format == DS_PAL) {
		init_params.format = DV1394_PAL;
	} else if (s->format == DS_NTSC) {
		init_params.format = DV1394_NTSC;
	} else {
		return NULL;
	}
		
	if (ioctl(s->fd, DV1394_IOC_INIT, &init_params) < 0) {
		fprintf(stderr, "Error: ioctl on the DV interface: %s\n", strerror(errno));
		return NULL;
	}
	
	s->ring = mmap(NULL, s->frame_len * RING_FRAMES, PROT_READ, MAP_PRIVATE, s->fd, 0);
	if (s->ring == MAP_FAILED) {
		fprintf(stderr, "Error: performing mmap for DV ring\n");
		return NULL;
	}	
	
	if (ioctl(s->fd, DV1394_IOC_START_RECEIVE, 0) < 0) {
		fprintf(stderr, "Error: ioctl on the DV interface: %s\n", strerror(errno));
		return NULL;
	}
	
	s->ring_current   = 0;
	s->ring_captured  = 0;
	s->ring_done      = 0;
	
	return s;
}

void
vidcap_dvlinux_done(void *state)
{
	struct vidcap_dvlinux_state *s;
	
	assert(state != NULL);
	
	s = (struct vidcap_dvlinux_state *) state;

	if ( s != NULL ) {
		
		if (ioctl(s->fd, DV1394_IOC_SHUTDOWN, 0) < 0) {
			fprintf(stderr, "Error: shutting down the DV device: %s.\n",
				strerror(errno));
		}
		
		if (munmap(s->ring, s->frame_len * RING_FRAMES) < 0) {
			fprintf(stderr, "Error: unmapping the DV ring: %s.\n",
				strerror(errno));
		}
		
		if (s->fd >= 0) {
			close(s->fd);
		}
		
		free(s);
	}
}

struct video_frame *
vidcap_dvlinux_grab(void *state)
{
	struct video_frame		*vf;
	struct vidcap_dvlinux_state	*s;
	struct pollfd			 poll_fd;
	int				 poll_result;
	struct dv1394_status		 status;


	assert(state != NULL);
	
	s = (struct vidcap_dvlinux_state *) state;
	
	assert(s->fd        > 0);
	assert(s->frame_len > 0);
	
	vf = (struct video_frame *) malloc(sizeof(struct video_frame));
	if (vf != NULL) {
			
		/* If we have no completed frames, request more */
		if(s->ring_captured <= 0) {
			
			/* Have we treated previous frames? Yes, then ask for more... */
			if (s->ring_done > 0) {
				if (ioctl(s->fd, DV1394_IOC_RECEIVE_FRAMES, s->ring_done) < 0) {
					debug_msg("DV capture ring overflow\n");
				}
			}

			/* Wait for frames in a poll */
			do {
				poll_fd.fd     = s->fd;
				poll_fd.events = POLLIN | POLLERR | POLLHUP;
				poll_result    = poll(&poll_fd, 1, -1);
				
				debug_msg("Waiting for more frames... (polling).\n");
			} while (poll_result < 0 && (errno == EAGAIN || errno == EINTR));
			
			if (poll_result < 0) {
				fprintf(stderr, "Error: in poll: %s.\n", strerror(errno));
				free(vf);
				return NULL;
			}

			/* Reset the ring counters, quering the status */
			if (ioctl(s->fd, DV1394_IOC_GET_STATUS, &status) < 0) {
				fprintf(stderr, "Error: ioctl on the DV interface: %s\n", strerror(errno));
			} else {
				s->ring_captured  = status.n_clear_frames;
				s->ring_current   = status.first_clear_frame;
				s->ring_done      = 0;
			
				if (status.dropped_frames) {
					debug_msg("WARNING! Dropped frames in the DV device.\n");
				}
			}
			
		}

//		printf("Current, Captured, Done = [%d,%d,%d]\n",
//		       s->ring_current,
//		       s->ring_captured,
//		       s->ring_done);
//		
//		debug_msg("Framelen=%d, %dx%d\n",
//			  s->frame_len, 
//			  display_get_width (s->format),
//			  display_get_height(s->format) );
		
		/* Create the frame */
		vf->colour_mode	= YUV_422;
		vf->width	= display_get_width (s->format);
		vf->height	= display_get_height(s->format);
		vf->data	= s->ring + (s->ring_current * s->frame_len);
		vf->data_len	= s->frame_len;
		
		/* Update pointers */
		s->ring_current = (s->ring_current + 1) % RING_FRAMES;
		s->ring_captured--;
		s->ring_done++;
        } else {
                fprintf(stderr,"Not enough space to allocate a video_frame.\n");
	}
	
	return vf;
}

struct vidcap_type *
vidcap_dvlinux_probe(const struct run_env* environ)
{
        struct vidcap_type		*vt	 = NULL;
	struct vidcap_dvlinux_state	*s	 = NULL;
	capture_format_t		*cformat = NULL;

        assert( environ != NULL);

	cformat = malloc(NUM_FORMATS * sizeof(capture_format_t));
	if(cformat != NULL) {
		cformat[0].size        = DS_PAL;
		cformat[0].colour_mode = YUV_422;
		cformat[0].fps         = 25;
		cformat[0].encoding    = dv;
		
		cformat[1].size        = DS_NTSC;
		cformat[1].colour_mode = YUV_422;
		cformat[1].fps         = 25;
		cformat[1].encoding    = dv;	
	} else {
		return NULL;
	}
	
	/*
	 * Start the environment and open the device
	 */
	s = vidcap_dvlinux_state_init( cformat, environ->capture_args );
	if( s->filename != NULL )
		s->fd  = open(s->filename, O_RDONLY);
	else
		dvlinux_open_default( s );
		
	if (s->fd >= 0) {		
		vt = (struct vidcap_type *) malloc(sizeof(struct vidcap_type));
		if (vt != NULL) {
			vt->id          = VIDCAP_DVLINUX_ID;
			vt->name        = "dv";
			vt->description = "IEEE-1394/DV";
			vt->props       = 0;
			vt->formats	= cformat;
			vt->num_formats = NUM_FORMATS;
		}
	}
	
	/* Restore previous state */
	vidcap_dvlinux_done(s);
	
        return vt;
}

#endif
