/*
 * FILE:   testfile.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.22 $
 * $Date: 2005/06/09 13:38:06 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_display.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_capture/testfile.h"
#include "video_codec.h"

#include "tv.h"

/* Mapping type */
struct ext_code_map_t {
        char			*ext;      
        vidcap_frame_enc_t	 decoder;  
};

/*
 * Mappings betweeen file name extensions and codecs
 */
struct ext_code_map_t     ext_table[] = {
	{".yuv", raw },		/* A YUV file is a raw movie */
	{".dv",  dv  }
};

#define EXT_TABLE_SIZE   (sizeof(ext_table) / sizeof(struct ext_code_map_t))

/*
 * State structure
 */ 
struct vidcap_testfile_state {
        int                          fd;
        struct testfile_movie_props *movie_props;

        char                        *base_ptr;    /* Start point in memory */
        char                        *read_ptr;    /* Current read pointer */
        long int                     len;         /* File size (in bytes) */

	struct timeval		     next_send_time;
};

/*
 * Some properties of the movie.
 */ 
struct testfile_movie_props {
        const char		*filename;	/* The name of the file		*/
        int			 fps;		/* Frames per second		*/ 
	display_size_t		 size;		/* Size o frames                */
	video_colour_mode_t	 color;		/* Color mode			*/
        long int		 frames;	/* Movie size (in frames)	*/
        unsigned int		 frame_bytes;	/* Size of each frame (in bytes)*/
	vidcap_frame_enc_t	 encoding;	/* Codec			*/  
};


/*
 * Find a codec given a filename, based on the file extension.
 */
static vidcap_frame_enc_t
find_codec(const char* in)
{
        unsigned int	i = 0, 
			j,
			k;
        int found = FALSE;
	
	assert(in != NULL);
	
        while (!found && i<EXT_TABLE_SIZE) {
                j = strlen(in) - 1;
                k = strlen(ext_table[i].ext) - 1;
		
                while (in[j] == ext_table[i].ext[k]) {
                        found = TRUE;
                        j--;
                        k--;
                }
		
                if(found && in[++j] == '.') {
			return ext_table[i].decoder;
		}
		
                found = FALSE;
                i++;
        }
	
        return raw;
}



/*
 * Parse the capture_args arguments and set dome other properties.
 * Now, capture_args==filename, but this could change. For example,
 * the command line argument could be "movie.yuv@25"
 */ 
static struct testfile_movie_props*
vidcap_testfile_props (capture_format_t *format, const char *args)
{
        struct testfile_movie_props	*movie_props;
	int				 temp_fd;
	unsigned			 temp_codec = VCODEC_INVALID_ID;
        
	UNUSED(format);
	UNUSED(args);
	
        movie_props = malloc(sizeof(struct testfile_movie_props));
        
        movie_props->filename = args;
	movie_props->encoding = raw;	
	
	/* If a filename has been provided, it must be valid */
        if (movie_props->filename != NULL) {
                if ((temp_fd = open( movie_props->filename, O_RDONLY))) {
			
			/* Find the codec for the movie format... */
			movie_props->encoding = find_codec(movie_props->filename);
				
                        close( temp_fd );
		} else {			
                        fprintf(stderr, "File [%s] open error: %s.\n", movie_props->filename, strerror(errno));
                        free(movie_props);
			return NULL;
		}
	}
	
	/* Get the frame size based on the codec used */
	if(movie_props->encoding == raw) {
		temp_codec = vcodec_get_by_name("uv_yuv422_8");
	} else {
		temp_codec = vcodec_get_by_enc(movie_props->encoding);
	}
			
	if (temp_codec != VCODEC_INVALID_ID) {
		movie_props->size	 = vcodec_decoder_get_size (temp_codec);
		movie_props->frame_bytes = vcodec_encoder_get_framelen (temp_codec);
		movie_props->color	 = vcodec_decoder_get_cm (temp_codec);
	} else {
		movie_props->size	 = DS_NONE;
		movie_props->frame_bytes = 0;
		movie_props->color	 = 0;		
	}
	
        movie_props->fps         = 0;	/* unknown */
        movie_props->frames      = 0;	/* known later */
        
        return movie_props;
}


/* External API ***********************************************************************************/

struct vidcap_type *
vidcap_testfile_probe(const struct run_env* environ)
{
        struct vidcap_type		*vt		= NULL;
        struct testfile_movie_props	*props		= NULL;
	capture_format_t		*cformat	= NULL;

	/* Delay the cformat initialization, until we have enought details */
	cformat = malloc(sizeof(capture_format_t));
	if (cformat == NULL) {
		fprintf(stderr, "Error: allocating memory for the list of capture formats.\n");
		return NULL;
	}
	
        props = vidcap_testfile_props( cformat, environ->capture_args );
        if( props != NULL ) {

		cformat[0].size        = props->size;
		cformat[0].colour_mode = props->color;
		cformat[0].fps         = props->fps;
		cformat[0].encoding    = props->encoding;
		
		vt = (struct vidcap_type *) malloc(sizeof(struct vidcap_type));
		if (vt != NULL) {
			vt->id          = VIDCAP_TESTFILE_ID;
			vt->name        = "testfile";
			vt->description = "Test File";
			vt->props       = VIDCAP_NEEDS_PARAMS |
				          VIDCAP_PARAM_FILENAME;
			vt->formats	= cformat;
			vt->num_formats = 1;
		} else {
			fprintf(stderr, "Not enough space when trying to allocate a vidcap_type.\n");
		}
        } else {
		fprintf(stderr, "Error retrieving the properties of the movie.\n");
	}
        
        return vt;
}

void *
vidcap_testfile_init(capture_format_t *format, const char *args)
{
        struct vidcap_testfile_state *s;
        int                           fd;
        struct testfile_movie_props  *props;
        struct stat                   statbuf;

	assert(format != NULL);

        props = vidcap_testfile_props( format, args );
        
        if( props->filename == NULL ) {
                fprintf(stderr, "The testfile capture needs a file name as a parameter! Use the -T flag.\n");
                return NULL;
        }        
     
        /*
         * Opens the file
         */
        if( (fd = open(props->filename, O_RDONLY)) >= 0 ) {

                /*
                 * Allocate space for the state struct
                 */ 
                s = (struct vidcap_testfile_state *)
                        malloc(sizeof(struct vidcap_testfile_state));
                if (s == NULL) {
                        fprintf(stderr,"Unable to allocate file state\n");
                        return NULL;
                }
                        
                /*
                 * Find size of input file
                 */
                if (fstat (fd, &statbuf) < 0) {
                        fprintf(stderr, "FSTAT error: %s.\n", strerror(errno) );
                        return NULL;
                }

                s->fd                  = fd;
                s->movie_props         = props;
                s->len                 = statbuf.st_size;
		
                s->movie_props->frames = s->len / s->movie_props->frame_bytes;
                
		s->base_ptr            = NULL;
                s->read_ptr            = NULL;
		
		gettimeofday(&(s->next_send_time), NULL);
		tv_add(&(s->next_send_time), 1.0/s->movie_props->fps);

		if( s->movie_props->frames == 0 ) {
                        fprintf(stderr, "The movie is shorter than a frame!!! Aborting...\n");
                        return NULL;
		}

                /*
                 * Print some information
                 */ 
                printf("Capture from file: %s\n", s->movie_props->filename);
                printf("              FPS: %d\n", s->movie_props->fps);
                printf("           Length: %ld frames, %ld seconds\n",
                        s->movie_props->frames,
                        (s->movie_props->frames / s->movie_props->fps));
		printf("            Codec: %s\n", 
		       vcodec_get_name(vcodec_get_by_enc(s->movie_props->encoding)) );

                if( (s->len % s->movie_props->frame_bytes) != 0 ) {
                        printf("WARNING: Movie length is not a multiple of the frame size!.\n");
                }
                
                /*
                 * mmap the input file
                 */
                if ((s->base_ptr = mmap (0, s->len, PROT_READ, 
                                    MAP_SHARED, s->fd, 0)) == (caddr_t) -1) {
                        fprintf(stderr, "MMAP error: %s.\n", strerror(errno) );
                        return NULL;
                }

#ifdef HAVE_MADVISE		
                madvise (s->base_ptr,  s->len, MADV_SEQUENTIAL);
#endif
                
                s->read_ptr = s->base_ptr;
        
                debug_msg("File capture device enabled\n");
                
                return s;
        };
        
        return NULL;        
}

void
vidcap_testfile_done(void *state)
{
        struct vidcap_testfile_state *s = (struct vidcap_testfile_state *) state;

        munmap(s->base_ptr, s->len);
        
        free(s);       
}

struct video_frame *
vidcap_testfile_grab(void *state)
{
        struct vidcap_testfile_state   *s = (struct vidcap_testfile_state *) state;
        struct video_frame             *vf;
        int                             m_frame_bytes;
        int                             m_offset;

	sleep_until(s->next_send_time);
	gettimeofday(&(s->next_send_time), NULL);
	tv_add(&(s->next_send_time), 1.0/(double) s->movie_props->fps);

        m_offset      = s->read_ptr - s->base_ptr;  /* Current offset in the movie */
        m_frame_bytes = s->movie_props->frame_bytes;

        /* 
         * If we are going to read out of the file, then loop
         */                   
        if ( (m_offset + m_frame_bytes) > s->len )
                s->read_ptr = s->base_ptr;

        vf = (struct video_frame *) malloc(sizeof(struct video_frame));
        if (vf != NULL) {
                vf->colour_mode = s->movie_props->color;
                vf->width       = display_get_width(s->movie_props->size);
                vf->height      = display_get_height(s->movie_props->size);
                vf->data        = (unsigned char*)s->read_ptr;
                vf->data_len    = m_frame_bytes;
        
                /* Update the read pointer */
                s->read_ptr    += m_frame_bytes; 
        } else
                fprintf(stderr,"Not enough space to allocate a video_frame.\n");
        
        return vf;
}

