/*
 * FILE:   video_codec.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.47 $
 * $Date: 2005/05/24 13:03:28 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "video_types.h"
#include "video_display.h"
#include "video_capture.h"
#include "rtp/rtp.h"		/* Needed by rtp/pbuf.h  */
#include "rtp/pbuf.h"		/* For struct coded_data */
#include "video_codec.h"
#include "video_codec/dv.h"
#include "video_codec/uv_yuv422_8.h"
#include "video_codec/uv_yuv422_10dvs.h"

struct vcodec_api {
	int                  (*vcx_init)            (void);
	void                 (*vcx_exit)            (void);
	const char *         (*vcx_get_name)        (void);
	const char *         (*vcx_get_desc)        (void);
	int                  (*vcx_can_encode)      (void);
	int                  (*vcx_can_decode)      (void);
	void                *(*vcx_encoder_create)  (unsigned id, unsigned mtu);
	void                 (*vcx_encoder_destroy) (void *state);
	unsigned int         (*vcx_encoder_get_framelen)(void);
	int                  (*vcx_encode)          (void *state, 
                                                     struct video_frame *in, 
					             struct encoded_frame *out,
						     vcodec_enc_opt_t opt);
	struct encoded_data *(*vcx_encode_get)      (struct encoded_frame *frame, 
						     int pos);
	video_colour_mode_t  (*vcx_decoder_get_cm)  (void);
	display_size_t       (*vcx_decoder_get_size)(void);
	unsigned int         (*vcx_decoder_get_framelen)(void);
	void                *(*vcx_decoder_create)  (unsigned id);
	void                 (*vcx_decoder_destroy) (void *state);
	int                  (*vcx_decode)          (void *state, 
                                                     struct coded_data *in, 
						     struct video_frame *out);
	int                  (*vcx_decode_raw)     (void *state, 
                                                     struct video_frame *in, 
						     struct video_frame *out);
	int                  (*vcx_decode_partial)  (void *state, 
                                                     struct coded_data *in, 
						     struct video_frame *out);
};

/* Table of pointers to codec functions, one for each codec we */
/* support. These are not necessarily all available for use as */
/* some codecs may fail their initialization (for example, if  */
/* they depend on hardware devices that are not present). The  */
/* available_vcodecs[] array contains pointers to those codecs */
/* that are usable.                                            */
static struct vcodec_api vcodec_table[] = {
#ifdef HAVE_LIBDV
	{
		dv_codec_init,
		dv_codec_exit,
		dv_codec_get_name,
		dv_codec_get_desc,
		dv_codec_can_encode,
		dv_codec_can_decode,
		dv_encoder_create,
		dv_encoder_destroy,
		dv_encoder_get_framelen,
		dv_encode,
		dv_encode_get,
		dv_decoder_get_cm,
		dv_decoder_get_size,		
		dv_decoder_get_framelen,
		dv_decoder_create,
		dv_decoder_destroy,
		dv_decode,
		dv_decode_raw,
		dv_decode_partial
	},
#endif
	{
		uv_yuv422_8_init,
		uv_yuv422_8_exit,
		uv_yuv422_8_get_name,
		uv_yuv422_8_get_desc,
		uv_yuv422_8_can_encode,
		uv_yuv422_8_can_decode,
		uv_yuv422_8_encoder_create,
		uv_yuv422_8_encoder_destroy,
		uv_yuv422_8_encoder_get_framelen,
		uv_yuv422_8_encode,
		uv_yuv422_8_encode_get,
		uv_yuv422_8_decoder_get_cm,
		uv_yuv422_8_decoder_get_size,
		uv_yuv422_8_decoder_get_framelen,
		uv_yuv422_8_decoder_create,
		uv_yuv422_8_decoder_destroy,
		uv_yuv422_8_decode,
		uv_yuv422_8_decode_raw,
		uv_yuv422_8_decode_partial
	},

        {
                uv_yuv422_10dvs_init,
                uv_yuv422_10dvs_exit,
                uv_yuv422_10dvs_get_name,
                uv_yuv422_10dvs_get_desc,
                uv_yuv422_10dvs_can_encode,
                uv_yuv422_10dvs_can_decode,
                uv_yuv422_10dvs_encoder_create,
                uv_yuv422_10dvs_encoder_destroy,
		uv_yuv422_10dvs_encoder_get_framelen,
                uv_yuv422_10dvs_encode,
                uv_yuv422_10dvs_encode_get,
                uv_yuv422_10dvs_decoder_get_cm,
                uv_yuv422_10dvs_decoder_get_size,
		uv_yuv422_10dvs_decoder_get_framelen,
                uv_yuv422_10dvs_decoder_create,
                uv_yuv422_10dvs_decoder_destroy,
                uv_yuv422_10dvs_decode,
		uv_yuv422_10dvs_decode_raw,
                uv_yuv422_10dvs_decode_partial
        }

};

#define NUM_VCODEC_INTERFACES (sizeof(vcodec_table)/sizeof(struct vcodec_api))

#define MAX_PAYLOAD 128

/* Pointers to entries in the vcodec_table[], for codecs that  */
/* are usable...                                               */
static struct vcodec_api *available_vcodecs[NUM_VCODEC_INTERFACES];
static unsigned	          num_vcodecs;
static unsigned           pt_map[MAX_PAYLOAD];


/*
 * Mappings between encodings of capture devices and codecs
 */
struct enc_cod_map_entry {
	vidcap_frame_enc_t	 enc;
	char			*codec;
};

static struct enc_cod_map_entry enc_cod_map_table[] = {
	{ raw,  ""   },
	{  dv,  "dv" }
};

#define NUM_ENC_COD_MAP (sizeof(enc_cod_map_table)/sizeof(struct enc_cod_map_entry))


/*
 * Public interface functions follow...
 */

void
vcodec_init(void)
{
	unsigned i;

	num_vcodecs = 0;
	for (i = 0; i < NUM_VCODEC_INTERFACES; i++) {
		if (vcodec_table[i].vcx_init()) {
			available_vcodecs[num_vcodecs++] = &vcodec_table[i];
		}
	}
	
	for (i = 0; i < MAX_PAYLOAD; i++) {
		/* Illegal - will trigger an assert() if unmapped codec is used*/
		pt_map[i] = VCODEC_INVALID_ID;	
	}
}

void
vcodec_done(void)
{
	unsigned i;

	for (i = 0; i < num_vcodecs; i++) {
		available_vcodecs[i]->vcx_exit();
	}
	num_vcodecs = 0;
}

unsigned
vcodec_get_num_codecs(void)
{
	return num_vcodecs;
}

const char *
vcodec_get_name(unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_get_name();
	} else {
		return "(invalid)";
	}
}

const char *
vcodec_get_description(unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_get_desc();
	} else {
		return NULL;
	}
}

uint8_t
vcodec_get_payload(unsigned id) {
	uint8_t pt = 0;
		
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		while ((pt_map[pt] != id) && (pt < MAX_PAYLOAD)) {
			pt++;
		}

		return pt;
	} else {
		return 0;
	}
}

int
vcodec_can_encode(unsigned id)
{	
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_can_encode();
	} else {
		return FALSE;
	}
}

int
vcodec_can_decode(unsigned id)
{	
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_can_decode();
	} else {
		return FALSE;
	}
}

void
vcodec_map_payload(uint8_t pt, unsigned id)
{
	assert(pt < MAX_PAYLOAD);
	assert(pt_map[pt] == VCODEC_INVALID_ID);	/* Fail if already mapped */
	pt_map[pt] = id;
}

void
vcodec_unmap_payload(uint8_t pt)
{
	assert(pt < MAX_PAYLOAD);
	assert(pt_map[pt] < num_vcodecs);		/* Fail if payload not mapped */
	pt_map[pt] = VCODEC_INVALID_ID;
}

unsigned
vcodec_get_by_payload(uint8_t pt)
{
	assert(pt < MAX_PAYLOAD);

	return pt_map[pt];
}

unsigned
vcodec_get_by_name(const char* name)
{
	unsigned int i;
	
	if (name != NULL)
		for (i = 0; i < vcodec_get_num_codecs(); i++)
			if (strcmp(vcodec_get_name(i), name) == 0)
				return i;

	return VCODEC_INVALID_ID;
}

unsigned
vcodec_get_by_enc (vidcap_frame_enc_t enc) {
	unsigned int i;

	for (i = 0; i < NUM_ENC_COD_MAP; i++)
		if(enc_cod_map_table[i].enc == enc)
			return vcodec_get_by_name(enc_cod_map_table[i].codec);
	
	return VCODEC_INVALID_ID;
}


/*****************************************************************************/

void *
vcodec_encoder_create(unsigned id, unsigned mtu)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_encoder_create(id,mtu);
	} else {
		return NULL;
	}
}

void
vcodec_encoder_destroy(unsigned id, void *state)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_destroy(state);
	}
}

unsigned
vcodec_encoder_get_framelen(unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_encoder_get_framelen();
	} else {
		return 0;
	}
}


int
vcodec_encode(unsigned              id,
              void                 *state, 
              struct video_frame   *in, 
              struct encoded_frame *out,
	      vcodec_enc_opt_t      opt)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_encode(state, in, out, opt);
	} else {
		return FALSE;
	}
}

struct encoded_data*
vcodec_encode_get(struct encoded_frame* frame, int pos)
{
	assert(frame != NULL);

	if (frame->id != VCODEC_INVALID_ID) {
		assert(frame->id < num_vcodecs);
				
		return available_vcodecs[frame->id]->vcx_encode_get(frame, pos);
	} else {
		return NULL;
	}
}

video_colour_mode_t
vcodec_decoder_get_cm(unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_get_cm();
	} else {
		return NONE;
	}
}

display_size_t
vcodec_decoder_get_size(unsigned id)
{	
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_get_size();
	} else {
		return DS_NONE;
	}
}

unsigned int
vcodec_decoder_get_framelen(unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_get_framelen();
	} else {
		return 0;
	}
}

void *
vcodec_decoder_create(unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_create(id);
	} else {
		return NULL;
	}
}

void
vcodec_decoder_destroy(unsigned id, void *state)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_destroy(state);
	}
}

int
vcodec_decode(unsigned             id,
              void                *state, 
              struct coded_data   *in, 
              struct video_frame  *out)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decode(state, in, out);
	} else {
		return FALSE;
	}
}

int
vcodec_decode_raw(unsigned   		 id,
		   void			*state,
		   struct video_frame	*in,
		   struct video_frame	*out)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decode_raw(state, in, out);
	} else {
		return FALSE;
	}
}

int
vcodec_decode_partial(unsigned             id,
                      void                *state, 
                      struct coded_data   *in, 
                      struct video_frame  *out)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decode_partial(state, in, out);
	} else {
		return FALSE;
	}
}

