/*
 * FILE:   video_codec.h
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.25 $
 * $Date: 2005/05/24 13:03:28 $
 *
 */

#ifndef _VCODEC_H
#define _VCODEC_H

typedef int vcodec_id_t;

/*
 * Encoding flags
 */
typedef uint32_t	vcodec_enc_opt_t;

#define VCODEC_NEED_PARAMS	0x0001	/* Needs initialization params */
#define VCODEC_ALREADY_ENCODED	0x0010	/* Data is already encoded */
#define VCODEC_FORMAT_PAL	0x0100	/* Use PAL format */
#define VCODEC_FORMAT_NTSC	0x0200	/* Use NTSC format */

/*
 * Initialisation, probing of available codecs, name and RTP 
 * payload type mapping functions. 
 */

#define VCODEC_INVALID_ID	0xffffffff

/*
 * The coded representation of a frame.
 * Used as a reference for the currently encoded/frame, but never as a
 * direct source for the encoded data. See encoded_data instead and its
 * associated functions.
 */
struct encoded_frame {
	uint32_t	 magic;
	unsigned	 id;		/* Codec ID			*/
	uint16_t	 len;		/* Number of nodes		*/
	unsigned long	 datalen;	/* Length of the encoded data	*/
	void		*priv;
};


/*
 * The coded representation of a fragment of a frame.
 */
struct encoded_data {
	uint16_t	 seqno;		/* Sequence number [0-n] */
	unsigned char	*header;	/* Header		 */
	uint16_t	 header_len;	/* Header lenght	 */
        unsigned char	*data;		/* Encoded data		 */
	uint16_t	 data_len;	/* Data lenght		 */
	uint16_t	 mark;
};

/*
 * Codecs management functions.
 */
void        vcodec_init(void);
void        vcodec_done(void);
unsigned    vcodec_get_num_codecs (void);

const char *vcodec_get_name       (unsigned id);	/* Single word name */
const char *vcodec_get_description(unsigned id);	/* Descriptive text */
uint8_t     vcodec_get_payload    (unsigned id);

int         vcodec_can_encode     (unsigned id);
int         vcodec_can_decode     (unsigned id);

void        vcodec_map_payload    (uint8_t pt, unsigned int id);
void        vcodec_unmap_payload  (uint8_t pt);
unsigned    vcodec_get_by_payload (uint8_t pt);
unsigned    vcodec_get_by_name    (const char* name);
unsigned    vcodec_get_by_enc     (vidcap_frame_enc_t enc);

/*
 * Video encoder and decoder functions. These operate on particular
 * instances of a codec, identified by the "state" parameter.
 */

void                *vcodec_encoder_create (unsigned id, unsigned mtu);
void                 vcodec_encoder_destroy(unsigned id, 
                                            void                 *state);
unsigned int         vcodec_encoder_get_framelen(unsigned id);
int                  vcodec_encode         (unsigned              id, 
                                            void                 *state,
                                            struct video_frame   *in,
					    struct encoded_frame *out,
					    vcodec_enc_opt_t      opt);

struct encoded_data *vcodec_encode_get      (struct encoded_frame*, int);

video_colour_mode_t  vcodec_decoder_get_cm  (unsigned id);
display_size_t       vcodec_decoder_get_size(unsigned id);
unsigned int         vcodec_decoder_get_framelen(unsigned id);
void                *vcodec_decoder_create  (unsigned id);
void                 vcodec_decoder_destroy (unsigned id, void   *state);
int                  vcodec_decode          (unsigned             id, 
                                             void                *state,
					     struct coded_data   *in,
                                             struct video_frame  *out);
int                  vcodec_decode_raw      (unsigned             id, 
                                             void                *state,
					     struct video_frame  *in,
                                             struct video_frame  *out);

int                  vcodec_decode_partial  (unsigned             id,
                                             void                *state,
					     struct coded_data   *in,
                                             struct video_frame  *out);

#endif /* _VCODEC_H */

