/*
 * FILE:   video_codec/dv.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.40 $
 * $Date: 2005/05/27 15:22:49 $
 *
 */

#include "config.h"
#include "config_unix.h"

#ifdef HAVE_LIBDV
#include "debug.h"
#include "rtp/rtp.h"
#include "video_types.h"
#include "video_display.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_codec.h"
#include "video_codec/dv.h"

#include "libdv/dv.h"

#define MAX_ENCODER_FRAMELEN	144000		/* Hardcoded for PAL */

/*
 * Encoder and decoder internal states.
 */
struct dv_enc_state {
	uint32_t		 magic;
	unsigned int		 id;

	dv_encoder_t		*encoder;

	display_size_t		 format;
	unsigned int		 mtu;
	
	struct encoded_data	*encoder_nodes;
	unsigned char		*encoder_buf;
};

struct dv_dec_state {
	uint32_t		magic;
	unsigned int		id;
	
	dv_decoder_t		*decoder;
	unsigned char		*decoder_buf;
	unsigned int		 decoder_buf_len;
};


int
dv_codec_init(void)
{
	return TRUE;	/* Codec always available */
}

void
dv_codec_exit(void)
{
	/* Nothing to do */
}

const char *
dv_codec_get_name(void)
{
	return "dv";
}

const char *
dv_codec_get_desc(void)
{
	return "DV (IEC 61834/SMPTE 314M)";
}

int
dv_codec_can_encode(void)
{
	return TRUE;
}

int
dv_codec_can_decode(void)
{
	return TRUE;
}

void *
dv_encoder_create(unsigned id, unsigned mtu)
{
	struct dv_enc_state	*s;
	unsigned int		 max_nodes = 0;
	
	s = (struct dv_enc_state*) malloc(sizeof(struct dv_enc_state));
	if (s == NULL) {
		fprintf(stderr, "Error: allocating memory for DV encoder state.\n");
		return NULL;
	}
	
	s->encoder   = NULL;
	s->id        = id;
	s->mtu       = mtu;
	s->format    = DS_PAL;

	/* Memory for an encoded frame */
	s->encoder_buf = (unsigned char*) malloc(MAX_ENCODER_FRAMELEN * sizeof(unsigned char));
	if (s->encoder_buf == NULL) {
		fprintf(stderr, "Error: allocating memory for a DV encoder frame.\n");
		dv_encoder_destroy(s);
		return NULL;
	}
	
	/* 
	 * Allocate memory for the nodes, estimating how many packets
	 * we will generate for a frame.
	 */
	max_nodes = (MAX_ENCODER_FRAMELEN / mtu) + 1;
	
	s->encoder_nodes = (struct encoded_data*) 
		malloc(max_nodes * sizeof(struct encoded_data));
	if (s->encoder_nodes == NULL) {
		fprintf(stderr, "Error: allocating memory for %d DV encoder nodes.\n", max_nodes);
		dv_encoder_destroy(s);
		return NULL;
	}
	
	return s;
}

void
dv_encoder_destroy(void *state)
{
	struct dv_enc_state *s = (struct dv_enc_state*) state;

	assert(s   != NULL);
	
	free(s->encoder_buf);
	free(s->encoder_nodes);
	free(s);
}

unsigned int
dv_encoder_get_framelen(void)
{
	return MAX_ENCODER_FRAMELEN;		/* Max size: PAL */
}

int
dv_encode(void                 *state,
	  struct video_frame   *in,
	  struct encoded_frame *out,
	  vcodec_enc_opt_t      opt)
{
	struct encoded_data	*nodes		= NULL;
	struct dv_enc_state	*s		= (struct dv_enc_state*) state;
	unsigned int		 framelen	= 0;
	unsigned int		 i		= 0;
	unsigned int		 data_left	= 0;
	unsigned char		*data_ptr	= NULL;
	unsigned int		 data_this_packet;

	assert(in  != NULL);
	assert(out != NULL);
	assert(s   != NULL);
	assert(s->mtu > 0);

	if (opt & VCODEC_ALREADY_ENCODED) {
		if(in->data_len > dv_encoder_get_framelen()) {
			fprintf(stderr, "WARNING: Encoded frame may not fit into the encoding pool.\n");
			fprintf(stderr, "         Encoded frame of size %d when only %d was expected.\n",
				in->data_len, dv_encoder_get_framelen());
		}

		data_ptr = in->data;
		framelen = in->data_len;
	} else {
		if(s->encoder == NULL) {
			/* Create a encoder */
		}

		printf("NOT IMPLEMENTED!!!\n");

		/* Encode the data, leaving the encoded frame in s->encoder_buf */
		data_ptr = s->encoder_buf;
		framelen = dv_encoder_get_framelen();   /* take the standar framelen */
	}

	/* Split the encoder buffer in nodes */
	nodes     = s->encoder_nodes;
	data_left = framelen;

	/* Frames always contain an integral number of 80 byte DIF blocks */
	assert((data_left % 80) == 0);

	while(data_left > 0) {		
		data_this_packet = (data_left < (s->mtu/80)*80 ? data_left : (s->mtu/80)*80);

		nodes[i].seqno		= i;
		nodes[i].header		= NULL;
		nodes[i].header_len	= 0;
		nodes[i].data		= data_ptr;
		nodes[i].data_len	= data_this_packet;

		data_left -= data_this_packet;
		data_ptr  += data_this_packet;
		
		i++;
	}

	/*
	 * Fill up the output
	 */
	out->id      = s->id;
	out->len     = i;
	out->datalen = framelen;
	out->magic   = s->magic;
	out->priv    = s->encoder_nodes;

	return out->len;	/* return the number of nodes generated */
}

struct encoded_data *
dv_encode_get(struct encoded_frame* frame, int pos)
{	
	struct encoded_data	*nodes;

	assert(frame       != NULL);
	assert(frame->priv != NULL);
	assert(pos         >= 0);

	if(pos < frame->len) {
		nodes = (struct encoded_data*) frame->priv;	
		return (nodes + pos);
	} else
		return NULL;
}

video_colour_mode_t
dv_decoder_get_cm(void)
{
	return YUV_422;
}

display_size_t
dv_decoder_get_size(void) {
	return DS_PAL;
};

unsigned int
dv_decoder_get_framelen(void)
{
	display_size_t	size = dv_decoder_get_size();
	return display_get_width(size) * display_get_height(size) * 2;	
}

void *
dv_decoder_create(unsigned id)
{
	struct dv_dec_state	*s;
	
	UNUSED(id);
	
	s = (struct dv_dec_state*) malloc(sizeof(struct dv_dec_state));
	if (s == NULL) {
		fprintf(stderr,"Error allocating memory for the DV decoder state.\n");
	} else {
		s->decoder = dv_decoder_new(FALSE, FALSE, FALSE);
		
		/* Create a buffe for storing the fragments of an encoded frame */
		s->decoder_buf_len = dv_encoder_get_framelen();
		s->decoder_buf = (unsigned char*) malloc(s->decoder_buf_len);
		if(s->decoder_buf == NULL) {
			debug_msg("Error: allocating memory for the DV decoder state.\n");
			dv_decoder_destroy(s);
		}

		s->decoder->quality = DV_QUALITY_BEST;
		s->decoder->video->quality = s->decoder->quality;
	}
		
	return (void*)s;
}

void
dv_decoder_destroy(void *state)
{
	struct dv_dec_state *s = (struct dv_dec_state*) state;

	assert( s          != NULL );
	assert( s->decoder != NULL );
	
	if (s->decoder_buf != NULL)
		free(s->decoder_buf);
		
	dv_decoder_free(s->decoder);
	
	free(s);
}

int
dv_decode(void *state, struct coded_data *in, struct video_frame *out)
{
	struct coded_data	*curr		= NULL;
	unsigned int		 i		= 0,
				 offset		= 0;
	int			 pitches[3];
	unsigned char		*pixels[3], tmp[4];
	struct dv_dec_state	*s = (struct dv_dec_state*) state;
	unsigned char		*pos;
	unsigned char		 sct, dseq, dbn;

	assert(s             != NULL);
	assert(s->decoder    != NULL);
	assert(out           != NULL);
	assert(out->data     != NULL);
	assert(out->data_len >= dv_decoder_get_framelen());

	/* Create the frame joining the fragments. Each RTP packet may */
	/* contain several DIF blocks.                                 */
	for (curr = in; curr != NULL; curr = curr->next) {
		if ((curr->data->meta.data_len % 80) != 0) {
			/* RFC 3189 requires the payload to contain an integral number */
			/* of 80 byte DIF blocks. Don't try to decode packets that are */
			/* the wrong size.                                             */
			debug_msg("RTP payload does not contain an integral number of DIF blocks (discarded)\n");
			continue;
		}

		pos = curr->data->meta.data;
		while (pos < (curr->data->meta.data + curr->data->meta.data_len)) {
			sct  = (unsigned) ((pos[0] & 0xe0) >> 5);
			dseq = (unsigned) ((pos[1] & 0xf0) >> 4);
			dbn  = (unsigned)  (pos[2]);

			switch (sct) {
			case 0: /* Header */
				offset = dseq * 12000;	
				break;
			case 1: /* Subcode */
				offset = dseq * 12000 + (dbn + 1) * 80;
				break;
			case 2: /* VAUX */
				offset = dseq * 12000 + (dbn + 3) * 80;
				break;
			case 3: /* Audio */
				offset = dseq * 12000 + 480 + (dbn * 16 * 80);
				break;
			case 4: /* Video */
				offset = dseq * 12000 + 480 + (dbn/15 * 16 * 80) + 80 + (dbn%15 * 80);
				break;
			default:
				debug_msg("Unknown DV frame type\n");
			}

			if ((offset + 80) <= s->decoder_buf_len) {
				memcpy(s->decoder_buf + offset, pos, 80);
			} else {
				debug_msg("DV frame offset out of bounds (%u > %u)\n", offset + 80, s->decoder_buf_len);
			}
			pos += 80;
		}
	}

	dv_parse_header(s->decoder, s->decoder_buf);
	dv_parse_packs(s->decoder, s->decoder_buf);

	pixels[0] = out->data;
	pixels[1] = 0;
	pixels[2] = 0;

	pitches[0] = 720 * 2;
	pitches[1] = 0;
	pitches[2] = 0;

	dv_decode_full_frame(s->decoder, s->decoder_buf, e_dv_color_yuv, pixels, pitches);

	/* The DV decoder gives us Y1UY2V colour component order, but we need UY1VY2. */
	/* Flip the data around to match the HDTV colour format used elsewhere...     */
	/* ****** FIXME: this is hardcoded for PAL image size!              [csp] */
	for (i = 0; i < 720 * 576; i += 2) {
		tmp[0] = out->data[(i * 2) + 0];
		tmp[1] = out->data[(i * 2) + 1];
		tmp[2] = out->data[(i * 2) + 2];
		tmp[3] = out->data[(i * 2) + 3];
#ifdef WORDS_BIGENDIAN
		out->data[(i * 2) + 3] = tmp[1];
		out->data[(i * 2) + 2] = tmp[0];
		out->data[(i * 2) + 1] = tmp[3];
		out->data[(i * 2) + 0] = tmp[2];
#else
		out->data[(i * 2) + 0] = tmp[1];
		out->data[(i * 2) + 1] = tmp[0];
		out->data[(i * 2) + 2] = tmp[3];
		out->data[(i * 2) + 3] = tmp[2];
#endif
	}

	out->width       = display_get_width (dv_decoder_get_size());
	out->height      = display_get_height(dv_decoder_get_size());
	out->colour_mode = dv_decoder_get_cm();
	out->data_len    = dv_decoder_get_framelen();	
	
	return 1;	
}

int
dv_decode_raw(void			*state,
	       struct video_frame	*in,
	       struct video_frame	*out)
{
	struct coded_data	new_in;
	rtp_packet              packet;

	packet.meta.data = in->data;
	packet.meta.data_len = in->data_len;

	new_in.seqno = 0;
	new_in.data  = &packet;
	new_in.next  = NULL;
	new_in.prev  = NULL;

	return dv_decode(state, &new_in, out);
}

int
dv_decode_partial(void *state, struct coded_data *in, struct video_frame *out)
{
	return dv_decode(state, in, out);
}

#endif /* HAVE_LIBDV */

