/*
 * FILE:   video_codec/uv_yuv422_10dvs_.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *         Ladan Gharai  <ladan@isi.edu>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 * Copyright (c) 2003-2004 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.10.2.2 $
 * $Date: 2005/08/21 00:10:23 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "video_types.h"
#include "video_display.h"
#include "video_capture.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_codec.h"
#include "video_codec/uv_yuv422_10dvs.h"

/* Bytes per pixel */
#define PGROUP 16  /* Bytes per PGROUP - per uncomp-draft */
#define PIXELS 6   /* Pixels per PGROUP - per uncomp-draft */

/* We limit packets to having up to 10 payload headers... */
#define MAX_PAYLOAD_COUNT 10    

#define END_OF_FRAME  (1<<15)


/* Payload header structure */
typedef struct {
	uint16_t        y_offset;       /* pixels */
	uint16_t        x_offset;       /* pixels */
	uint16_t        length;         /* pixels */
	uint16_t        flags;          
} payload_hdr_t;

/* Internal state */
struct vcodec_yuv422_10dvs_state {
	uint32_t		 magic;
	unsigned int		 mtu;
	unsigned int		 id;

	unsigned int		 pool_len;
	struct encoded_data	*pool_node;	/* Pool for nodes */	
	payload_hdr_t		*pool_head;	/* Pool for payload headers */ 
};

/******************************************************************************
 * Public API
 *****************************************************************************/

int
uv_yuv422_10dvs_init(void)
{
	return 1;	/* Codec always available */
}

void
uv_yuv422_10dvs_exit(void)
{
	/* Nothing to do */
}

const char *
uv_yuv422_10dvs_get_name(void)
{
	return "uv_yuv422_10";
}

const char *
uv_yuv422_10dvs_get_desc(void)
{
	return "Uncompressed YUV 4:2:2 (10 bit)";
}

int
uv_yuv422_10dvs_can_encode(void)
{
	return TRUE;
}

int
uv_yuv422_10dvs_can_decode(void)
{
	return TRUE;
}

void *
uv_yuv422_10dvs_encoder_create(unsigned id, unsigned mtu)
{
	struct vcodec_yuv422_10dvs_state	*state;
	unsigned int			 num_nodes,
					 max_node_len,
					 max_head_len;
	
	state = (struct vcodec_yuv422_10dvs_state*)
		malloc(sizeof(struct vcodec_yuv422_10dvs_state));
	
	if( state == NULL )
		fprintf(stderr, "Allocation error for the YUV 4:2:2 10bit DVS codec state.");
	else {
		state->magic = VCODEC_YUV422_10DVS_ID;
		state->mtu   = mtu;
		state->id    = id;		
		
		/*
		 * Estimate the number of packets to generate.
		 */
		max_node_len = 2488320; //((720*1280*2)/3)*4;  /* subtitute accordingly */ 
		max_head_len = 8 * MAX_PAYLOAD_COUNT;		
		
		num_nodes = max_node_len * 4 / (mtu - 40 - max_head_len);
		
		state->pool_len = num_nodes;
		
		/*
		 * Allocate the three pools
		 */
		state->pool_node = (struct encoded_data*) 
			malloc(num_nodes * sizeof(struct encoded_data));
		if(state->pool_node == NULL) {
			fprintf(stderr, "Error allocating memory for %d nodes in the encoder pool.\n",
				num_nodes);
			uv_yuv422_10dvs_encoder_destroy(state);
		}
				
		state->pool_head = (payload_hdr_t*)
			malloc(num_nodes * MAX_PAYLOAD_COUNT * sizeof(payload_hdr_t));
		if(state->pool_head == NULL) {
			fprintf(stderr, "Error allocating memory for %d headers in the encoder pool.\n",
				num_nodes * MAX_PAYLOAD_COUNT);
			uv_yuv422_10dvs_encoder_destroy(state);
		}
		
		debug_msg("Codec memory pool created for %d nodes.\n", num_nodes);
	}
	
	return (void*) state;
}

void
uv_yuv422_10dvs_encoder_destroy(void *state)
{
	struct vcodec_yuv422_10dvs_state *s = (struct vcodec_yuv422_10dvs_state*) state;
	
	assert( s != NULL );
	
	if (s->pool_node != NULL)
		free(s->pool_node);
	if (s->pool_head != NULL)
		free(s->pool_head);
	
	free(s);
}

unsigned int            
uv_yuv422_10dvs_encoder_get_framelen(void)
{
        /* it doesn't have a real meaning in this codec */
        display_size_t  size;
                        
        size = uv_yuv422_10dvs_decoder_get_size();
                                
        return size;            
}           

int
uv_yuv422_10dvs_encode(void *state, struct video_frame *in, struct encoded_frame *out, vcodec_enc_opt_t opt)
{	
	unsigned int			 mark		= 0,
					 x		= 0,
					 y		= 0,
					 first_x	= 0,
					 first_y	= 0,
					 inc =0 ,
					 data_len	= 0,
					 payload_count	= 0,
					 node_count	= 0,
					 funny_counter  = 0,
					 no_pgroups,
					 pgroups_left_this_line = 0,
					 pgroups_left_this_node;
	
	payload_hdr_t			*curr_head;
	struct encoded_data		*nodes;
		
	struct vcodec_yuv422_10dvs_state	*s = (struct vcodec_yuv422_10dvs_state*) state;
	
	UNUSED(opt);
	
	assert (out		!= NULL);
	assert (s		!= NULL);
	assert (s->magic	== VCODEC_YUV422_10DVS_ID);
	assert (in		!= NULL);
	assert (in->colour_mode == uv_yuv422_10dvs_decoder_get_cm());
	debug_msg ("%d\n", in->data_len );
	assert (in->data_len    == 2488320 ); 
	
	curr_head = s->pool_head;
	nodes     = s->pool_node;

	if (in->width==1280) {in->width+=4;}	/* HDs and CENT */

	do {    /* remnant */
		/* figure out how many groups of 6 pixels are left 
		switch (funny_counter) {
		case 0: pgroups_left_this_line = ((in->width +4 - x) / PIXELS); break;
		case 1: pgroups_left_this_line = ((in->width +2 - x) / PIXELS); break;
		case 2: pgroups_left_this_line = ((in->width  - x) / PIXELS); break;
		default : assert (funny_counter > 2 );
		}*/
		pgroups_left_this_line = ((in->width - x) / PIXELS); 
		/* figure out how many PGROUPS are left */
		pgroups_left_this_node = (s->mtu - 40 - 
			data_len - (8 * (payload_count + 1))) / PGROUP;
		
		/* Calculate the length of this node */
		if (pgroups_left_this_node < pgroups_left_this_line)
			no_pgroups = pgroups_left_this_node;
		else
			no_pgroups = pgroups_left_this_line;
		
		curr_head[payload_count].y_offset = htons(y);
		curr_head[payload_count].x_offset = htons(x);
		curr_head[payload_count].length   = htons(no_pgroups*PGROUP);
		curr_head[payload_count].flags    = htons(0);
		payload_count++;
		//printf ("x:%4d y:%3d %5d \n", x, y, no_pgroups*PGROUP );
		
		data_len = data_len + no_pgroups*PGROUP;
		
		x += no_pgroups* PIXELS; /* number of pixels sent */

		/* If we finish this line, reset X and increment Y... 
                switch (funny_counter) {
                case 0: if (x==in->width+4) {x=4; y++; funny_counter++;}; break;
                case 1: if (x==in->width+2) {x=2; y++; funny_counter++;}; break;
                case 2: if (x==in->width)   {x=0; y++; funny_counter=0;}; break;
                default : assert (funny_counter > 2 );
                }*/
		if (x==in->width) {x=0; y++;};
		
		if (y == in->height)
			mark = 1;
		
                /* Is it time to build a packet? */
		if ((y == in->height)     || 
		    (payload_count == MAX_PAYLOAD_COUNT) || 
		    ((40u + data_len + (8u * (payload_count + 1)) + PGROUP) > s->mtu)) {
			
			/* Mark the end */
			curr_head[payload_count-1].flags = htons(END_OF_FRAME);
				
			/* Fill up the node */
			nodes[node_count].header     = (unsigned char*) curr_head;
			nodes[node_count].header_len = 8 * payload_count;
			
			
			nodes[node_count].data       = in->data + 
						((first_y *in->width +first_x+inc ) / PIXELS) * PGROUP;
			assert (((first_y*in->width + first_x) % PIXELS )==0);	
			nodes[node_count].data_len   = data_len;
			nodes[node_count].seqno      = node_count;
			nodes[node_count].mark       = mark;

			/* Update pointers */
			curr_head		+= MAX_PAYLOAD_COUNT;
			
			/* Reset counters */
			payload_count		 = 0;
			node_count		+= 1;
			data_len		 = 0;
			first_x			 = x;
			//inc=12; /* remnant */
			first_y			 = y;
			
			//printf ("%d %d\n", node_count, s->pool_len );
			assert(node_count < s->pool_len);

		}
	} while (y < in->height);

	/* Create the encoded frame */
	out->magic = VCODEC_YUV422_10DVS_ID;
	out->id    = s->id;
	out->priv  = (void*) s->pool_node;
	out->len   = node_count;

	return node_count;
}

struct encoded_data *
uv_yuv422_10dvs_encode_get(struct encoded_frame* frame, int pos)
{
	struct encoded_data	*nodes;
	
	assert(frame		!= NULL);
	assert(frame->priv	!= NULL);

	nodes = (struct encoded_data*) frame->priv;
	
	if( pos < frame->len )
		return (nodes + pos);
	else
		return NULL;
}

video_colour_mode_t 
uv_yuv422_10dvs_decoder_get_cm(void)
{
	return YUV_422_DVS10;
}

display_size_t 
uv_yuv422_10dvs_decoder_get_size(void)
{
	return DS_720p;
}

unsigned int         
uv_yuv422_10dvs_decoder_get_framelen(void) {
	display_size_t	size;
	
	size = uv_yuv422_10dvs_decoder_get_size();
	
	return 2488320;
}


void *
uv_yuv422_10dvs_decoder_create(unsigned id)
{
	struct vcodec_yuv422_10dvs_state	*state;
	
	state = (struct vcodec_yuv422_10dvs_state*) malloc(sizeof(struct vcodec_yuv422_10dvs_state));
	if( state == NULL )
		fprintf(stderr, "Allocation error for the YUV 4:2:2 codec state.");
	else {
		state->magic = VCODEC_YUV422_10DVS_ID;
		state->mtu   = 0;
		state->id    = id;		
	}	
		
	return state;
}

void
uv_yuv422_10dvs_decoder_destroy(void *state)
{
	uv_yuv422_10dvs_encoder_destroy(state);
}


int
uv_yuv422_10dvs_decode(void *state, struct coded_data *in, struct video_frame *out)
{
	/* Copy 1 rtp packet to frame for uncompressed HDTV data. */
	payload_hdr_t			*curr_hdr;
	payload_hdr_t			*hdr[MAX_PAYLOAD_COUNT];
	unsigned int			 hdr_count	= 0,
					 frame_offset	= 0,
					 inc = 0,
					 i;
	unsigned char			*base,
					*offset;	
	struct vcodec_yuv422_10dvs_state	*s = (struct vcodec_yuv422_10dvs_state*) state;

 	assert(in               != NULL);
	assert(out		!= NULL);
	assert(out->colour_mode == uv_yuv422_10dvs_decoder_get_cm());
	assert(out->data        != NULL);
	assert(s		!= NULL);
	assert(s->magic		== VCODEC_YUV422_10DVS_ID );

	if (out->width==1280) {out->width+=4;} /* HDs and CENT */
		
	while (in != NULL) {
		
		assert(in->data            != NULL) ;
		assert(in->data->meta.data != NULL);

		/* Figure out how many headers... */
		curr_hdr = (payload_hdr_t *) in->data->meta.data;
		while (TRUE) {
			hdr[hdr_count++] = curr_hdr;
			if ((ntohs(curr_hdr->flags) & END_OF_FRAME) != 0) {
				/* Last header... */
				break;
			}
			if (hdr_count == 10) {
				/* Out of space... */
				break;
			}
			curr_hdr++;
		}
		
		/* OK, now we can copy the data */
		offset = (in->data->meta.data) + hdr_count * 8;
		for (i = 1; i < hdr_count + 1; i++) { 
			unsigned short x_offset = ntohs(hdr[i - 1]->x_offset);
			unsigned short y_offset = ntohs(hdr[i - 1]->y_offset);
			unsigned short length   = ntohs(hdr[i - 1]->length);
 	
			if ((x_offset > out->width) || (y_offset > out->height)) {
				debug_msg("Coded frame out of bounds\n");
				return 0;
			}

			/* remnant */
                        //if ((x_offset==0) && (y_offset==0)) {inc=0;}
                        //if ((x_offset==0) && (y_offset!=0)) {inc=12; x_offset+=inc;}
                        //printf ("%d %d\n", x_offset, inc);

			frame_offset = ((x_offset +(y_offset * out->width))/PIXELS)*PGROUP;
			
			/* Check that the memcpy can be done */
			if (frame_offset + length > out->data_len) {
				printf ("%d %d %d\n", length, out->data_len, out->colour_mode );
				fprintf(stderr, "Encoded data does not fit into the frame. Skiping...\n");
				return 1;
			} else {
				/* Ok, move the data to the frame */
				base = out->data + frame_offset;
				memcpy(base,offset,length);
				offset+=length;
			}
		}
		
		in           = in->next;
		hdr_count    = 0;
		frame_offset = 0;
	}

	return 1;
}

int
uv_yuv422_10dvs_decode_partial(void *state, struct coded_data *in, struct video_frame *out)
{
	return uv_yuv422_10dvs_decode(state, in, out);
}

int
uv_yuv422_10dvs_decode_raw(void *state, struct video_frame *in, struct video_frame *out)
{
	UNUSED(state);
	
	out = in;
	
	return 1;
}

