/*
 * FILE:   video_display/dumpfile.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.9 $
 * $Date: 2005/06/23 21:41:55 $
 *
 */


#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_display.h"

#include "video_display/dumpfile.h"

#define HD_WIDTH    1280
#define HD_HEIGHT   720

#define MAGIC_DUMPFILE   DISPLAY_DUMPFILE_ID

/*
 * Internal state
 */
struct state_dumpfile {
	int		fd;
	
	char		*base_ptr;	/* Start point in memory */
	char		*write_ptr;	/* Current write pointer */	
	unsigned long	len;		/* Lenght of the movie */
	
	struct video_frame		frame;	
	struct dumpfile_movie_props	*movie_props;

	/* For debugging... */
	uint32_t        magic;
};

/*
 * Some properties of the movie
 */ 
struct dumpfile_movie_props {
        char    *filename;         /* The name of the file */
        int      width;            /* Width and...  */
        int      height;           /* height of the movie */
        long int frames;           /* Movie size (in frames), currently not used */
        int      frame_bytes;      /* Size of each frame (in bytes) */
};

/*
 * Parse the display_args arguments and set some other properties.
 * By now, display_args=="filename@frames". This could change. 
 * For example, the command line argument could be "movie.yuv@25@640x480"
 */ 
static struct dumpfile_movie_props*
dumpfile_props(char* display_args)
{
	struct dumpfile_movie_props*	movie_props;
	char*				token;

	movie_props = malloc(sizeof(struct dumpfile_movie_props));
	if( movie_props == NULL ) {
		fprintf(stderr, "Unable to allocate memory for the dumpfile command line options parsing.\n");
		return NULL;
	}

	/*
	 * By now, parsing is very easy...
	 */ 
	movie_props->filename = (char*)strtok(display_args, "@");
	
	token = strtok(NULL, "@");
	if (token) {
		movie_props->frames = atol(token);
	} else
		movie_props->frames = 0;

	movie_props->width    = HD_WIDTH;   /* ignored */
	movie_props->height   = HD_HEIGHT;  /* ignored */

	/*
	 * This is hardcoded for YUV
	 * FIXME: set this depending on the movie format detected/specified.
	 */ 
	movie_props->frame_bytes   = movie_props->width * movie_props->height * 2;
        
	return movie_props;
}



/**********************************************************************************
 * Public API
 *********************************************************************************/ 


/*
 * Dump-to-File output initialization.
 */
void*
display_dumpfile_init(char* display_args)
{
	int				fd;
	struct state_dumpfile		*s;
	struct dumpfile_movie_props	*props;

	s = malloc(sizeof(struct state_dumpfile));
	if( s == NULL ) {
		fprintf(stderr, "Unable to allocate memory for the dumpfile state.\n");
		return NULL;
	}

	s->magic = MAGIC_DUMPFILE;
        
	props = dumpfile_props( display_args );	 
	if( props == NULL ) {
		fprintf(stderr, "Unable to parse the command line options for the dumpfile device.\n");
		return NULL;
	}        
        
	if( props->filename == NULL ) {
		fprintf(stderr, "The dumpfile display needs a file name as a parameter! Use the -D flag.\n");
		return NULL;
	}

	if( s->movie_props->frames == 0 ) {
		fprintf(stderr, "Please specify the length of the dumpfile movie using the format -D filename@frames.\n");
		return NULL;
	}

	/*
	 * Opens the file
	 */
	if ((fd = open(props->filename, O_WRONLY)) >= 0)
	{	
		s = (struct state_dumpfile *) malloc(sizeof(struct state_dumpfile));
		if (s == NULL) {
			fprintf(stderr, "Unable to allocate memory for the dump-to-file internal state.\n");
			return NULL;
		}
		
		s->fd = fd;
		s->movie_props = props;
		s->len = s->movie_props->frame_bytes * s->movie_props->frames;
				
		s->base_ptr	= NULL;
		s->write_ptr	= NULL;

		printf("Dump-to-file: %s [ %ld frames, %ld bytes]",
			s->movie_props->filename, s->movie_props->frames, s->len);
				
		/*
		 * mmap the output file
		 */
		if ((s->base_ptr = mmap(0, s->len, PROT_WRITE,
					MAP_FILE, s->fd, 0)) == (caddr_t) - 1) {
			fprintf(stderr, "MMAP error for dump file: %s.\n", strerror(errno));
			return NULL;
		}
		
#ifdef HAVE_MADVISE
		madvise(s->base_ptr, s->len, MADV_SEQUENTIAL);
#endif	

		s->write_ptr = s->base_ptr;
		
		/*
		 * Creates a frame with the expected information
		 */
		s->frame.colour_mode	= YUV_422;
		s->frame.height		= s->movie_props->height;
		s->frame.width		= s->movie_props->width;
		s->frame.data_len	= s->movie_props->frame_bytes;		
		
		return s;
	} else
		perror("Opening the dumpfile:");
	
	return NULL;
}

void 
display_dumpfile_done(void *state)
{
	struct state_dumpfile *s = (struct state_dumpfile *) state;

	assert(s != NULL);
	assert(s->magic == MAGIC_DUMPFILE);

	/*
	 * Close the file
	 */
	munmap(s->base_ptr, s->len);

	free(s);       	 
}

struct video_frame *
display_dumpfile_getf(void *state, display_size_t size, video_colour_mode_t color)
{
	struct state_dumpfile *s = (struct state_dumpfile *) state;

	assert(s != NULL);
	assert(s->magic == MAGIC_DUMPFILE);

	UNUSED( size );
	UNUSED( color );
	
	/*
	 * The rest of the fields should be intact...
	 */
	s->frame.data		= (unsigned char*)s->write_ptr;

	return &(s->frame);	/* returns the frame */
}

display_result_t 
display_dumpfile_putf(void *state, struct video_frame * frame)
{
	struct state_dumpfile *s = (struct state_dumpfile *) state;

	assert(s != NULL);
	assert(s->magic == MAGIC_DUMPFILE);
	assert(frame != NULL);

	/*
	 * Update the writing pointer if we have enough space,
	 * otherwise, do nothing, and it will overwrite the last frame (ups!)
	 */
	if( s->write_ptr + frame->data_len < s->base_ptr + s->len )	
		s->write_ptr += frame->data_len;
	
	return DISPLAY_OK;
}

/*
 * Color mode suppported by the dump-to-file device
 */
video_colour_mode_t 
display_dumpfile_colour(void *state)
{
	struct state_dumpfile *s = (struct state_dumpfile *) state;

	assert(s != NULL);
	assert(s->magic == MAGIC_DUMPFILE);

	return YUV_422;
}

/*
 * Video formats supported by the dump-to-file device
 */
display_type_t *
display_dumpfile_probe(const struct run_env* environ, char* display_args)
{
	int				 temp_fd;
	display_type_t			*dt		= NULL;
	display_format_t		*dformat	= NULL;
	struct dumpfile_movie_props	*props;

     	UNUSED (environ);  
	props = dumpfile_props( display_args );

	/*
	 * If there is no filename provided, then this capture device 
	 * is not available...
	 */ 
	if( props->filename != NULL )
		if( (temp_fd = open( props->filename, O_WRONLY)) ) {
        
			dformat = malloc(4 * sizeof(display_format_t));
			dformat[0].size = DS_QCIF;
			dformat[0].colour_mode = YUV_422;
			dformat[0].num_images = 1;
			dformat[1].size = DS_CIF;
			dformat[1].colour_mode = YUV_422;
			dformat[1].num_images = 1;
			dformat[2].size = DS_SCIF;
			dformat[2].colour_mode = YUV_422;
			dformat[2].num_images = 1;
			dformat[3].size = DS_720p;
			dformat[3].colour_mode = YUV_422;
			dformat[3].num_images = 1;
		
			dt = malloc(sizeof(display_type_t));
			if (dt != NULL) {
				dt->id = DISPLAY_DUMPFILE_ID;
				dt->name = "dumpfile";
				dt->description = "Dump-to-File";
				dt->formats = dformat;
				dt->num_formats = 4;
			} else
				fprintf(stderr, "Memory allocation failed for the dump-to-file probe.\n");

                        close( temp_fd );
		} 
		
	return dt;
}

