/*
 * FILE:   video_display/hdstation.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *         Ladan Gharai  <ladan@isi.edu>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 * Copyright (c) 2001-2005 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.37 $
 * $Date: 2005/06/24 17:07:14 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_display.h"
#include "video_display/hdstation.h"
#include "tv.h"

#ifdef HAVE_HDSTATION		/* From config.h */

#include "dvs_clib.h"		/* From the DVS SDK */
#include "dvs_fifo.h"		/* From the DVS SDK */

typedef unsigned long uintptr;  /* FIX ME: place this where it should be */

video_colour_mode_t mode=4;     /* this is the part where I cheated - although */
                                /* it is possible to add a var to state?  -LG- */



#define HDSP_MAGIC	0x12345678

struct state_hdsp {
	pthread_t		 thread_id;
	sv_handle 		*sv;
	sv_fifo	 		*fifo;
	struct video_frame	 frame;
        sv_fifo_configinfo       config;  /* the centaurus needs alignment */
	sv_fifo_buffer		*fifo_buffer;
	sv_fifo_buffer		*display_buffer;
	sv_fifo_buffer		*tmp_buffer;
	pthread_mutex_t		 lock;
	pthread_cond_t		 boss_cv;
	pthread_cond_t		 worker_cv;
	int			 work_to_do;
	int			 boss_waiting;
	int			 worker_waiting;
	uint32_t		 magic;
	char			*bufs[2];
	char                    *tmp_buf[2];
	int			 bufs_index;
};

static void*
display_thread_hd(void *arg)
{
	struct state_hdsp       *s = (struct state_hdsp *) arg;
	int			 res;

	while (1) {
		pthread_mutex_lock(&s->lock);

		while (s->work_to_do == FALSE) {
			s->worker_waiting = TRUE;
			pthread_cond_wait(&s->worker_cv, &s->lock);
			s->worker_waiting = FALSE;
		}

		s->display_buffer = s->tmp_buffer;
		s->work_to_do     = FALSE;

		if (s->boss_waiting) {
			pthread_cond_signal(&s->boss_cv);
		}
		pthread_mutex_unlock(&s->lock);

		res = sv_fifo_putbuffer(s->sv, s->fifo, s->display_buffer, NULL);
		if (res != SV_OK) {
			debug_msg("Error %s\n", sv_geterrortext(res));
			return NULL;
		}
	}
	return NULL;
}

struct video_frame *
display_hdstation_getf(void *state, display_size_t size, video_colour_mode_t color)
{
	struct state_hdsp	*s = (struct state_hdsp *) state;
	int			 res;
	
	assert(s->magic == HDSP_MAGIC);
	
	UNUSED(size);
	//UNUSED(color);

	/* Prepare the new RTP buffer... */
	res = sv_fifo_getbuffer(s->sv, s->fifo, &s->fifo_buffer, NULL, SV_FIFO_FLAG_VIDEOONLY | SV_FIFO_FLAG_FLUSH);
	if (res != SV_OK) {
		debug_msg("Error %s\n", sv_geterrortext(res));
		return NULL;
	}
	s->bufs_index   = (s->bufs_index + 1) % 2;
	s->frame.data   = s->bufs[s->bufs_index];
	if (color==YUV_422)
		s->frame.data_len = 1280 * 720 * 2;
	else
		s->frame.data_len = 2488320; /*(1280 * 720 * 20 ) / 8 */

	s->fifo_buffer->dma.addr = s->frame.data;
	s->fifo_buffer->dma.size = s->frame.data_len;

	return &(s->frame);
}

display_result_t
display_hdstation_putf(void *state, struct video_frame *frame)
{
	struct state_hdsp	*s = (struct state_hdsp *) state;

	UNUSED(frame);

	assert(s->magic == HDSP_MAGIC);

	pthread_mutex_lock(&s->lock);
	/* Wait for the worker to finish... */
	while (s->work_to_do) {
		s->boss_waiting = TRUE;
		pthread_cond_wait(&s->boss_cv, &s->lock);
		s->boss_waiting = FALSE;
	}

	/* ...and give it more to do... */
	s->tmp_buffer  = s->fifo_buffer;
	s->fifo_buffer = NULL;
	s->work_to_do  = TRUE;

	/* ...and signal the worker */
	if (s->worker_waiting) {
		pthread_cond_signal(&s->worker_cv);
	}
	pthread_mutex_unlock(&s->lock);

	return DISPLAY_OK;
}

void *
display_hdstation_init(display_format_t *format, char* display_args)
{
	struct state_hdsp 	*s;
	int 		 	 res;
	
	UNUSED(display_args);

	/* Start the display thread... */
	s = (struct state_hdsp *) malloc(sizeof(struct state_hdsp));
	s->magic        = HDSP_MAGIC;

	s->sv  = sv_open("");
	if (s->sv == NULL) {
		debug_msg("Cannot open HDTV display device\n");
		return NULL;
	}

	if (format->colour_mode==YUV_422)
		res = sv_videomode(s->sv, SV_MODE_SMPTE296_60P | SV_MODE_STORAGE_FRAME );
	else
		res = sv_videomode(s->sv, SV_MODE_SMPTE296_60P | SV_MODE_STORAGE_FRAME | SV_MODE_NBIT_10B);

	if (res != SV_OK) {
		debug_msg("Cannot set videomode %s\n", sv_geterrortext(res));
	        return NULL;
	}
	sleep (1.2);

        res = sv_fifo_init(s->sv, &s->fifo, 0, 1, 1, 0, 0);
        if (res != SV_OK) {
                debug_msg("Cannot initialize video display FIFO %s\n", sv_geterrortext(res));
                return NULL;
        }

        res = sv_sync_output(s->sv, SV_SYNCOUT_DEFAULT | SV_SYNCOUT_OUTPUT_GREEN|SV_SYNC_EXTERNAL);	
	if (res != SV_OK) {
		debug_msg("Cannot enable sync-on-green %s\n", sv_geterrortext(res));
	        return NULL;
	}


        res = sv_fifo_configstatus(s->sv, s->fifo, &s->config);
        if(res != SV_OK)  {
                printf("sv_fifo_configstatus failed %d '%s'", res, sv_geterrortext(res));
		return NULL;
        }

	res = sv_fifo_start(s->sv, s->fifo);
	if (res != SV_OK) {
		debug_msg("Cannot start video display FIFO  %s\n", sv_geterrortext(res));
	        return NULL;
	}

	pthread_mutex_init(&s->lock, NULL);
	pthread_cond_init(&s->boss_cv, NULL);
	pthread_cond_init(&s->worker_cv, NULL);
	s->work_to_do     = FALSE;
	s->boss_waiting   = FALSE;
	s->worker_waiting = FALSE;
	s->display_buffer = NULL;

	/* allocate buffers */
        s->tmp_buf[0] = malloc(s->config.vbuffersize*8 + s->config.abuffersize + (s->config.dmaalignment-1));
        s->bufs[0] = (char *)((uintptr)(s->tmp_buf[0] + (s->config.dmaalignment-1)) & ~(uintptr)(s->config.dmaalignment-1));
        s->tmp_buf[1] = malloc(s->config.vbuffersize*8 + s->config.abuffersize + (s->config.dmaalignment-1));
        s->bufs[1] = (char *)((uintptr)(s->tmp_buf[1] + (s->config.dmaalignment-1)) & ~(uintptr)(s->config.dmaalignment-1));
	
        printf ("Video buffer:%d %d\n", s->config.vbuffersize, s->config.dmaalignment-1 );
	s->bufs_index = 0;

	if (pthread_create(&(s->thread_id), NULL, display_thread_hd, (void *) s) != 0) {
		perror("Unable to create display thread\n");
		return NULL;
	}

	s->frame.colour_mode = format->colour_mode; 
	s->frame.width       = 1280;
	s->frame.height      = 720;
	s->frame.data        = NULL;
	s->frame.data_len    = 0;

	return (void *) s;
}

void
display_hdstation_done(void *state)
{
	struct state_hdsp *s = (struct state_hdsp *) state;

	sv_fifo_free(s->sv, s->fifo);
	sv_close(s->sv);
	free(s);
}

video_colour_mode_t
display_hdstation_colour(void *state)
{
	struct state_hdsp *s = (struct state_hdsp *) state;

	assert(s->magic == HDSP_MAGIC);
	assert( mode < 4);

	return mode;  /* FIX ME !!!!!!!! - LG */
}

char	dvs_display_dev_name[50];

display_type_t *
display_hdstation_probe(const struct run_env* environ, char* display_args)
{
	display_type_t		*dtype;
	display_format_t	*dformat;
	sv_handle		*sv;
	video_colour_mode_t     colour_mode=4;
	int                     res;

	
	UNUSED(display_args);

	/* Probe the hardware... */
	sv  = sv_open("");
	if (sv == NULL) {
		debug_msg("Cannot probe HDTV display device\n");
		return NULL;
	}


        if (strcmp (environ->codec, "uv_yuv422_8")==0)  /* --LG-- */
                colour_mode = YUV_422;
        else
                if (strcmp (environ->codec, "uv_yuv422_10")==0)
                        colour_mode = YUV_422_DVS10;
	/* perhaps should use sizeof(video_colour_mode_t) here ?*/
        assert (colour_mode<4); 

        if (colour_mode == YUV_422)
		res = sv_videomode(sv, SV_MODE_SMPTE296_60P | SV_MODE_COLOR_YUV422 | SV_MODE_AUDIO_NOAUDIO);
	else	
	        res = sv_videomode(sv, SV_MODE_SMPTE296_60P | SV_MODE_NBIT_10B | 
							SV_MODE_COLOR_YUV422 | SV_MODE_AUDIO_NOAUDIO);
        if (res!=SV_OK) {
		debug_msg("Cannot set videomode\n");
	        return NULL;
	}
	sleep (1.2);
	sv_close(sv);

	dformat = malloc(sizeof(display_format_t));
	if (dformat == NULL) {
		return NULL;
	}
	dformat->size        = DS_720p;
	dformat->colour_mode = colour_mode;
	dformat->num_images  = 1;

	sprintf(dvs_display_dev_name, "DVS HDstationOEM/Centaurus [sdk v%d.%d.%d]", 
	             DVS_VERSION_MAJOR, DVS_VERSION_MINOR, DVS_VERSION_PATCH);

	dtype = malloc(sizeof(display_type_t));
	if (dtype != NULL) {
		dtype->id	   = DISPLAY_HDSTATION_ID;
		dtype->name        = "hdtv";
		dtype->description = dvs_display_dev_name;
		dtype->formats     = dformat;
		dtype->num_formats = 1;
	}
	return dtype;
}

#endif

