/*
 * FILE:   video_display/sdl.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.26 $
 * $Date: 2005/06/24 21:32:21 $
 *
 */


#include "config.h"
#include "config_unix.h"

#ifdef HAVE_LIBSDL

#include <SDL.h>  

#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_display.h"
#include "video_display/sdl.h"
#include "tv.h"

#define MAGIC_SDL   DISPLAY_SDL_ID

struct state_sdl {
    /* Internal state */
    SDL_Surface		*screen;   
    SDL_Overlay		*yuv_overlay;
    
    display_size_t	size;	/* current display size */
    video_colour_mode_t color;  /* current display color mode */ 
   
    uint8_t		video_bpp;

    struct video_frame  frame;

    int			init;
    
    /* For debugging... */
    uint32_t		magic;
};



/*
 * Set the SDL display.
 * The size and/or the color are expected to be different to the current values.
 */
static display_result_t
sdl_reconfigure(struct state_sdl* state, display_size_t size, video_colour_mode_t color)
{
	const SDL_VideoInfo *vi;
	char                 driver_name[128];
	int                  flags = SDL_SWSURFACE;
	//int		     flags = SDL_DOUBLEBUF|SDL_HWSURFACE;
	
	assert (size         != DS_NONE);
	assert (state        != NULL);
	assert (state->magic == MAGIC_SDL);
	
	if (state->init == FALSE) {
		if (SDL_Init(SDL_INIT_VIDEO) == -1) {
			fprintf(stderr, "Error: initializing the SDL.\n");
			display_sdl_done(state);
			return DISPLAY_ERR;
		}

		/*
		 * Some cosmetic stuff...
		 */ 
		SDL_ShowCursor( 1 );
		SDL_WM_SetCaption("UltraGrid", NULL);
		
		state->init = TRUE;
	}
	
	/*
	 * Do we need to resize the display?
	 */
	if (state->screen == NULL || state->size != size) {

		debug_msg("Setting display size for %dx%d.\n", 
			  display_get_width(size), display_get_height(size));
		
		state->screen = SDL_SetVideoMode( display_get_width(size),
						  display_get_height(size),
						  state->video_bpp,
						  flags);
		if(state->screen == NULL) {
			fprintf(stderr, "Unable to set video: %s\n", SDL_GetError());
			return DISPLAY_ERR;
		}
		
		/* 
		 * SDL video info
		 */
		vi = SDL_GetVideoInfo();
		SDL_VideoDriverName(driver_name, sizeof(driver_name));
		debug_msg("Video output:\n");
		debug_msg(" Driver: sdl (SDL driver: %s)\n", driver_name);
		debug_msg(" HW surface=%s\n", (vi->hw_available ? "yes" : "no"));
		debug_msg(" Window Manager=%s\n", (vi->wm_available ? "yes" : "no"));
		debug_msg(" BLIT accelaration: %s\n", (vi->blit_hw ? "yes" : "no"));
		debug_msg(" HWSW=%s\n", (vi->blit_sw ? "yes" : "no"));
		debug_msg(" Video Memory: %dKB\n", vi->video_mem);
		debug_msg(" Preferred Depth: %d bits/pixel\n", vi->vfmt->BitsPerPixel);
		debug_msg(" Must lock=%s\n", (SDL_MUSTLOCK(state->screen) ? "yes" : "no"));

		state->size = size;   /* Update the display id */
	}
	
	if (state->color != YUV_422 && color == YUV_422) {

		debug_msg("Creating YUV overlay.\n");
		
		/*
		 * Creates a YUV overlay for the display
		 */ 
		state->yuv_overlay = 
			SDL_CreateYUVOverlay(display_get_width(state->size),
					     display_get_height(state->size),
					     SDL_UYVY_OVERLAY,
					     state->screen);
		
		if ( state->yuv_overlay == NULL ) {
			fprintf(stderr, "SDL: Couldn't create SDL YUV overlay: %s\n",
				SDL_GetError());
			return DISPLAY_ERR;
		}
		
		/*
		 * Some debug info
		 */ 
		debug_msg(" YUV Overlay size: %dx%d \n", 
			  state->yuv_overlay->w, state->yuv_overlay->h);
		debug_msg(" YUV hardware acceleration: %s\n", 
			  ( state->yuv_overlay->hw_overlay == 1 ? "yes" : "no"));
		
		/*
		 * Displays a initial black frame
		 */ 
		SDL_DisplayYUVOverlay(state->yuv_overlay,
				      &(state->screen->clip_rect) );
	
		state->color = YUV_422;
		
	} else {
		debug_msg("sdl_reconfigure() called with an uknown color mode.");
	}

	return DISPLAY_OK;
};

/***************************************************************************
 * Public API
 **************************************************************************/

/*
 * SDL output initialization.
 */
void*
display_sdl_init(display_format_t *format, const char* display_args)
{
    struct state_sdl    *sdl_display;
    
    UNUSED( display_args );

    sdl_display = malloc(sizeof(struct state_sdl));

    sdl_display->magic     = MAGIC_SDL;
    sdl_display->init      = FALSE;
    sdl_display->video_bpp = 0;
    sdl_display->screen    = NULL;
    sdl_display->size      = DS_NONE;     /* Set the initial display size */
    sdl_display->color     = 0;
    
    /* We can force XV output setting this... */ 
#ifdef HAVE_TCL_TK
    setenv("SDL_VIDEO_X11_NODIRECTCOLOR", "1", 1);  
#endif
    setenv("SDL_VIDEO_YUV_HWACCEL", "1", 1);
            
    debug_msg("Delaying display initialization.\n");
        
    return sdl_display;
}

void
display_sdl_done(void *state)
{
    struct state_sdl *s = (struct state_sdl *) state;

    assert( s != NULL );
    assert( s->magic == MAGIC_SDL );

    /* Free the state */
    free(s);
    
    /* Clean up the SDL library */
    if (SDL_WasInit(SDL_INIT_VIDEO))
	    SDL_QuitSubSystem(SDL_INIT_VIDEO);
}

struct
video_frame *display_sdl_getf(void *state, display_size_t size, video_colour_mode_t color)
{
    struct state_sdl *sdl_state = (struct state_sdl *) state;

    assert( sdl_state        != NULL );
    assert( sdl_state->magic == MAGIC_SDL );

    /*
     * Reconfigure the display for the frame we will return.
     * If it is the first time we give a frame, we can not delay the SDL 
     * initialization after this point...
     */
    if( sdl_state->init   == FALSE || 
	sdl_state->screen == NULL  ||
	size != sdl_state->size    || 
	color != sdl_state->color  ) {
	    
	    sdl_reconfigure( sdl_state, size, color );
    }
    
    assert(sdl_state->screen != NULL);
    
    /* If we need it, lock the screen */
    if (SDL_MUSTLOCK(sdl_state->screen)) {
       if (SDL_LockSurface(sdl_state->screen) < 0) {
          fprintf(stderr, "Error: locking output screen: %s.\n", SDL_GetError());
          return NULL;
       }
    }
    
    if (SDL_LockYUVOverlay(sdl_state->yuv_overlay) < 0) {
	    fprintf(stderr, "Error: locking YUV overlay: %s.\n", SDL_GetError());
	    return NULL;
    }

    /*
     * Creates a frame with the expected size, and the data
     * section is the overlay structure...
     */ 
    sdl_state->frame.colour_mode = YUV_422;
    sdl_state->frame.height      = display_get_height(sdl_state->size);
    sdl_state->frame.width       = display_get_width(sdl_state->size);
    sdl_state->frame.data_len    = sdl_state->frame.height * sdl_state->frame.width * 2;
    sdl_state->frame.data        = sdl_state->yuv_overlay->pixels[0];    
    
    return &(sdl_state->frame);   /* returns the frame */
}

display_result_t
display_sdl_putf(void *state, struct video_frame *frame)
{
    struct state_sdl  *s = (struct state_sdl *) state;
    SDL_Event          event;

    assert( s        != NULL );
    assert( s->magic == MAGIC_SDL );
    assert( s->init  == TRUE );
    assert( frame    != NULL );
    
    if (SDL_MUSTLOCK(s->screen))
        SDL_UnlockSurface(s->screen);

    SDL_UnlockYUVOverlay(s->yuv_overlay);   
   
    SDL_DisplayYUVOverlay(s->yuv_overlay, &(s->screen->clip_rect));
    
    /* Control the exit status */
    SDL_PollEvent(&event);
    if ( event.type == SDL_QUIT )
        return DISPLAY_EXIT_REQ;
    
    if ( event.type == SDL_KEYDOWN &&
	 event.key.keysym.sym == SDLK_ESCAPE )
        return DISPLAY_EXIT_REQ;
    
    return DISPLAY_OK;
}

/*
 * Color mode suppported by the SDL
 */
video_colour_mode_t
display_sdl_colour(void *state)
{
    struct state_sdl *s = (struct state_sdl *) state;

    assert( s != NULL );
    assert( s->magic == MAGIC_SDL );

    return YUV_422;
}

/*
 * Video formats supported by the SDL
 */
display_type_t *
display_sdl_probe(const struct run_env* environ, const char* display_args)
{
    display_type_t *dt;
    display_format_t *dformat;

    UNUSED( display_args );
    UNUSED( environ );

    dformat = malloc(4 * sizeof(display_format_t));
    dformat[0].size = DS_QCIF;
    dformat[0].colour_mode = YUV_422;
    dformat[0].num_images = 1;
    dformat[1].size = DS_CIF;
    dformat[1].colour_mode = YUV_422;
    dformat[1].num_images = 1;
    dformat[2].size = DS_SCIF;
    dformat[2].colour_mode = YUV_422;
    dformat[2].num_images = 1;
    dformat[3].size = DS_720p;
    dformat[3].colour_mode = YUV_422;
    dformat[3].num_images = 1;

    dt = malloc(sizeof(display_type_t));
    if (dt != NULL) {
	dt->id = DISPLAY_SDL_ID;
	dt->name = "sdl";
	dt->description = "SDL";
	dt->formats = dformat;
	dt->num_formats = 4;
    }
    return dt;
}

#endif
