/*
 * FILE:   video_display/xv.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 * Copyright (c) 2001-2003 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.28 $
 * $Date: 2005/06/23 22:24:42 $
 *
 */

#include "config.h"
#include "config_unix.h"

#ifndef X_DISPLAY_MISSING /* Don't try to compile if X is not present */
#ifdef HAVE_TCL_TK        /* Need Tcl/Tk currently                    */

#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_display.h"
#include "video_display/xv.h"
#include "tcl.h"
#include "tk.h"
#include "tv.h"

/* For X shared memory... */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>

#define HD_WIDTH 	1280
#define HD_HEIGHT 	720
#define MAGIC_XV	DISPLAY_XV_ID

extern Tcl_Interp      *interp;		/* FIXME */

struct state_xv {
	Tk_Window		 vw;
	Display			*display;
	Window			 window;
	GC			 gc;
	int			 vw_depth;
	Visual			*vw_visual;
	XvImage			*vw_image[2];
	XShmSegmentInfo		 vw_shm_segment[2];
	int			 image_display, image_network;
	XvAdaptorInfo		*ai;
	int			 xv_port;
	struct video_frame	 frame;
	/* Thread related information follows... */
	pthread_t		 thread_id;
	pthread_mutex_t		 lock;
	pthread_cond_t		 boss_cv;
	pthread_cond_t		 worker_cv;
	int			 work_to_do;
	int			 boss_waiting;
	int			 worker_waiting;
	/* For debugging... */
	uint32_t		 magic;	
};

static void*
display_thread_xv(void *arg)
{
	struct state_xv        *s = (struct state_xv *) arg;

	while (1) {
		pthread_mutex_lock(&s->lock);

		while (s->work_to_do == FALSE) {
			s->worker_waiting = TRUE;
			pthread_cond_wait(&s->worker_cv, &s->lock);
			s->worker_waiting = FALSE;
		}

		s->work_to_do     = FALSE;

		if (s->boss_waiting) {
			pthread_cond_signal(&s->boss_cv);
		}
		pthread_mutex_unlock(&s->lock);

		XvShmPutImage(s->display, s->xv_port, s->window, s->gc, s->vw_image[s->image_display], 0, 0, 
			      s->vw_image[s->image_display]->width, s->vw_image[s->image_display]->height, 0, 0, HD_WIDTH, HD_HEIGHT, False);
		while (Tcl_DoOneEvent(TCL_ALL_EVENTS | TCL_DONT_WAIT)) {
			/* Process Tcl/Tk events */
		}
	}
	return NULL;
}


void *
display_xv_init(display_format_t *format, const char* display_args)
{
	struct state_xv		*s;
	unsigned int		 p_version, p_release, p_request_base, p_event_base, p_error_base, p; 
	int                   	 p_num_adaptors, i, j;
	
	UNUSED( display_args );
	UNUSED( format );

	s = (struct state_xv *) malloc(sizeof(struct state_xv));
	s->magic   = MAGIC_XV;
	s->xv_port = -1;

	/* Create a bare window to draw into... */
	s->vw = Tk_CreateWindowFromPath(interp, Tk_MainWindow(interp), ".vw", 0);
	if (s->vw == 0) {
		printf("Xv: Cannot create video window\n");
		return NULL;
	};
	Tk_SetClass(s->vw, "hdrecv_video");
	Tk_GeometryRequest(s->vw, HD_WIDTH, HD_HEIGHT);
	Tk_MapWindow(s->vw);
	s->vw_depth  = Tk_Depth(s->vw);
	s->vw_visual = Tk_Visual(s->vw);

	/* For now, we only support 24-bit TrueColor displays... */
	if (s->vw_depth != 24) {
		printf("Xv: Display doesn't support 24 bit colour\n");
		return NULL;
	}
	if (s->vw_visual->class != TrueColor) {
		printf("Xv: Display doesn't support true colour visual mode\n");
		return NULL;
	}

	/* Do we support the Xv extension? */
	if (XvQueryExtension(Tk_Display(s->vw), &p_version, &p_release, &p_request_base, &p_event_base, &p_error_base) != Success) {
		printf("Xv: Display doesn't support the X video extension\n");
		return NULL;
	}
	if (XvQueryAdaptors(Tk_Display(s->vw), DefaultRootWindow(Tk_Display(s->vw)), &p_num_adaptors, &s->ai) != Success) {
		printf("Xv: Display doesn't support the X video extension (cannot query adaptors)\n");
		return NULL;
	}
	if (p_num_adaptors == 0) {
		printf("Xv: Display doesn't support the X video extension (no display adaptors)\n");
		return NULL;
	}

	for (i = 0; i < p_num_adaptors; i++) {
		debug_msg("Xv: adaptor %d name = %s\n", i, s->ai[i].name);
		debug_msg("Xv: adaptor %d type = %s%s%s%s%s\n", i, (s->ai[i].type & XvInputMask)   ? "[input] "  : "",
						  		  (s->ai[i].type & XvOutputMask)  ? "[output] " : "",
								  (s->ai[i].type & XvVideoMask)   ? "[video] "  : "",
								  (s->ai[i].type & XvStillMask)   ? "[still] "  : "",
								  (s->ai[i].type & XvImageMask)   ? "[image] "  : "");
		s->xv_port = s->ai[i].base_id;
		for (p = s->ai[i].base_id; p < s->ai[i].base_id + s->ai[i].num_ports; p++) {
			int			 encodings, formats;
			XvEncodingInfo		*ei;
			XvImageFormatValues	*fo;

			if (XvQueryEncodings(Tk_Display(s->vw), p, &encodings, &ei) != Success) {
				printf("Xv: Cannot query video encodings\n");
				return NULL;
			}
			for (j = 0; j < encodings; j++) {
				debug_msg("Xv adaptor %d port %d coding %d = %s\n", i, p - s->ai[i].base_id, j, ei[j].name);
			}
			XvFreeEncodingInfo(ei);
			fo = XvListImageFormats(Tk_Display(s->vw), p, &formats);
			for (j = 0; j < formats; j++) {
				debug_msg("Xv adaptor %d port %d format %d = 0x%08lx %s\n", i, p - s->ai[i].base_id, j, fo[j].id, fo[j].guid);
			}
			if (fo != NULL) {
				XFree(fo);
			}
		}
	}

	/* Create the image buffer, shared with the X server... */
	for (i = 0; i < 2; i++) {
		s->vw_image[i] = XvShmCreateImage(Tk_Display(s->vw), s->xv_port, 0x59565955, 0, HD_WIDTH, HD_HEIGHT, &s->vw_shm_segment[i]);
		if (s->vw_image[i] == NULL) {
			printf("Xv: Cannot create shared memory image\n");
			goto failure;
		}
		debug_msg("vw_image                   = %p\n", s->vw_image[i]);
		debug_msg("vw_image->width            = %d\n", s->vw_image[i]->width);
		debug_msg("vw_image->height           = %d\n", s->vw_image[i]->height);
		debug_msg("vw_image->data_size        = %d\n", s->vw_image[i]->data_size);
		if (s->vw_image[i]->width != HD_WIDTH) {
			printf("Xv: Display does not support %d pixel wide Xvideo shared memory images\n", HD_WIDTH);
			goto failure;
		}
		if (s->vw_image[i]->height != HD_HEIGHT) {
			printf("Xv: Display does not support %d pixel tall Xvideo shared memory images\n", HD_WIDTH);
			goto failure;
		}

		s->vw_shm_segment[i].shmid    = shmget(IPC_PRIVATE, s->vw_image[i]->data_size, IPC_CREAT|0777);
		s->vw_shm_segment[i].shmaddr  = shmat(s->vw_shm_segment[i].shmid, 0, 0);
		s->vw_shm_segment[i].readOnly = False;
		debug_msg("vw_shm_segment.shmid       = %d\n", s->vw_shm_segment[i].shmid);
		debug_msg("vw_shm_segment.shmaddr     = %d\n", s->vw_shm_segment[i].shmaddr);

		s->vw_image[i]->data = s->vw_shm_segment[i].shmaddr;

		if (XShmAttach(Tk_Display(s->vw), &s->vw_shm_segment[i]) == 0) {
			printf("Xv: Cannot attach shared memory segment\n");
			goto failure;
		}
	}
	s->image_network = 0;
	s->image_display = 1;

	s->frame.colour_mode = YUV_422;
	s->frame.width       = HD_WIDTH;
	s->frame.height      = HD_HEIGHT;
	s->frame.data        = s->vw_image[0]->data;
	s->frame.data_len    = HD_WIDTH * HD_HEIGHT * 2;

	/* Get our window onto the screen... */
	Tcl_Eval(interp, "pack .vw");

	s->display = Tk_Display(s->vw);
	s->window  = Tk_WindowId(s->vw);
	s->gc      = Tk_GetGC(s->vw, 0, 0);

	XSync(s->display, False);

	pthread_mutex_init(&s->lock, NULL);
	pthread_cond_init(&s->boss_cv, NULL);
	pthread_cond_init(&s->worker_cv, NULL);
	s->work_to_do     = FALSE;
	s->boss_waiting   = FALSE;
	s->worker_waiting = TRUE;
	if (pthread_create(&(s->thread_id), NULL, display_thread_xv, (void *) s) != 0) {
		perror("Unable to create display thread\n");
		return NULL;
	}

	return (void *) s;

failure:
	/* Something went wrong during initialization... Try to cleanup: */
	for (i = 0; i < 2; i++) {
		if (s->vw_image[i] != NULL) {
			XShmDetach(Tk_Display(s->vw), &(s->vw_shm_segment[i]));
			shmdt(s->vw_shm_segment[i].shmaddr);
			shmctl(s->vw_shm_segment[i].shmid, IPC_RMID, 0);
			/*XDestroyImage(s->vw_image[i]);*/
		}
	}
	XvFreeAdaptorInfo(s->ai);
	return NULL;
}

void
display_xv_done(void *state)
{
	int		 i;
	struct state_xv *s = (struct state_xv *) state;

	assert(s->magic == MAGIC_XV);

	for (i = 0; i < 2; i++) {
		XShmDetach(Tk_Display(s->vw), &(s->vw_shm_segment[i]));
		shmdt(s->vw_shm_segment[i].shmaddr);
		shmctl(s->vw_shm_segment[i].shmid, IPC_RMID, 0);
		/*XDestroyImage(s->vw_image[i]);*/
	}
	XvFreeAdaptorInfo(s->ai);
}

struct video_frame *
display_xv_getf(void *state, display_size_t size, video_colour_mode_t color)
{
	struct state_xv *s = (struct state_xv *) state;
	assert(s->magic == MAGIC_XV);
	s->frame.data = s->vw_image[s->image_network]->data;
	return &(s->frame);
}

display_result_t
display_xv_putf(void *state, struct video_frame *frame)
{
	int		 tmp;
	struct state_xv *s = (struct state_xv *) state;

	assert(s->magic == MAGIC_XV);
	assert(frame->data == (unsigned char *) s->vw_image[s->image_network]->data);

	pthread_mutex_lock(&s->lock);
	/* Wait for the worker to finish... */
	while (s->work_to_do) {
		s->boss_waiting = TRUE;
		pthread_cond_wait(&s->boss_cv, &s->lock);
		s->boss_waiting = FALSE;
	}

	/* ...and give it more to do... */
	tmp = s->image_display;
	s->image_display = s->image_network;
	s->image_network = tmp;
	s->work_to_do    = TRUE;

	/* ...and signal the worker */
	if (s->worker_waiting) {
		pthread_cond_signal(&s->worker_cv);
	}
	pthread_mutex_unlock(&s->lock);
	return DISPLAY_OK;
}

video_colour_mode_t
display_xv_colour(void *state)
{
	struct state_xv *s = (struct state_xv *) state;
	assert(s->magic == MAGIC_XV);
	return YUV_422;
}

display_type_t *
display_xv_probe(const struct run_env* environ, const char* display_args)
{
        display_type_t          *dt;
        display_format_t        *dformat;
        
        UNUSED( display_args );
	UNUSED( environ );

	dformat = malloc(4 * sizeof(display_format_t));
	dformat[0].size        = DS_QCIF;
        dformat[0].colour_mode = YUV_422;
        dformat[0].num_images  = 1;
	dformat[1].size        = DS_CIF;
        dformat[1].colour_mode = YUV_422;
        dformat[1].num_images  = 1;
	dformat[2].size        = DS_SCIF;
        dformat[2].colour_mode = YUV_422;
        dformat[2].num_images  = 1;
	dformat[3].size        = DS_720p;
        dformat[3].colour_mode = YUV_422;
        dformat[3].num_images  = 1;

        dt = malloc(sizeof(display_type_t));
        if (dt != NULL) {
		dt->id	        = DISPLAY_XV_ID;
		dt->name        = "xv";
		dt->description = "X Window System with Xvideo extension";
                dt->formats     = dformat;
                dt->num_formats = 4;
        }
        return dt;
}

#endif /* HAVE_TCL_TK       */
#endif /* X_DISPLAY_MISSING */

